# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from collections.abc import Iterator, Sequence
from typing import Any, Callable, Optional

import fiona
import fiona.crs
import fsspec
import pyproj
import shapely
import shapely.ops

from xcube.server.api import ApiError, Context
from xcube.webapi.common.context import ResourcesContext

PlaceGroup = dict[str, Any]
Feature = dict[str, Any]

ALL_PLACES = "all"


class PlacesContext(ResourcesContext):
    _feature_index: int = 0

    def __init__(self, server_ctx: Context):
        super().__init__(server_ctx)
        self._additional_place_groups: dict[str, list[PlaceGroup]] = dict()
        self._place_group_cache: dict[str, PlaceGroup] = dict()

    def on_dispose(self):
        if self._place_group_cache:
            self._place_group_cache.clear()

    def get_cached_place_group(self, place_group_id: str) -> Optional[PlaceGroup]:
        return self._place_group_cache.get(place_group_id)

    def set_cached_place_group(self, place_group_id: str, place_group: PlaceGroup):
        self._place_group_cache[place_group_id] = place_group

    def get_cached_place_groups(
        self, predicate: Optional[Callable[[str, PlaceGroup], bool]] = None
    ) -> list[PlaceGroup]:
        return [
            v
            for k, v in self._place_group_cache.items()
            if predicate is None or predicate(k, v)
        ]

    def new_feature_id(self) -> str:
        with self.rlock:
            feature_id = str(self._feature_index)
            self._feature_index += 1
            return feature_id

    def get_global_place_groups(
        self, base_url: str, load_features=False
    ) -> list[PlaceGroup]:
        return self.load_place_groups(
            self.config.get("PlaceGroups", []),
            base_url,
            is_global=True,
            load_features=load_features,
        )

    def get_global_place_group(
        self, place_group_id: str, base_url: str, load_features: bool = False
    ) -> PlaceGroup:
        place_group_config = self._get_place_group_config(place_group_id)
        return self._load_place_group(
            place_group_config, base_url, is_global=True, load_features=load_features
        )

    def _get_place_group_config(self, place_group_id: str) -> dict:
        place_group_configs = self.config.get("PlaceGroups", [])
        for place_group_config in place_group_configs:
            if place_group_config["Identifier"] == place_group_id:
                return place_group_config
        raise ApiError.NotFound(f'Place group "{place_group_id}" not found')

    def load_place_groups(
        self,
        place_group_configs: list,
        base_url: str,
        is_global: bool = False,
        load_features: bool = False,
        qualifiers: Optional[list[str]] = None,
    ) -> list[PlaceGroup]:
        place_groups = []
        for place_group_config in place_group_configs:
            place_group = self._load_place_group(
                place_group_config,
                base_url,
                is_global=is_global,
                load_features=load_features,
            )
            place_groups.append(place_group)
        for q in [
            qualifier
            for qualifier in (qualifiers or ())
            if qualifier in self._additional_place_groups
        ]:
            for place_group in self._additional_place_groups[q]:
                place_groups.append(place_group)
        return place_groups

    def add_place_group(
        self, place_group: PlaceGroup, qualifiers: Optional[list[str]] = None
    ):
        for qualifier in qualifiers or ():
            if qualifier not in self._additional_place_groups:
                self._additional_place_groups[qualifier] = []
            self._additional_place_groups[qualifier].append(place_group)

    def _load_place_group(
        self,
        place_group_config: dict[str, Any],
        base_url: str,
        is_global: bool = False,
        load_features: bool = False,
    ) -> PlaceGroup:
        place_group_id = place_group_config.get("PlaceGroupRef")
        if place_group_id:
            if is_global:
                raise ApiError.InvalidServerConfig(
                    "'PlaceGroupRef' cannot be used in a global place group"
                )
            if len(place_group_config) > 1:
                raise ApiError.InvalidServerConfig(
                    "'PlaceGroupRef' if present,"
                    " must be the only entry in a 'PlaceGroups' item"
                )
            return self.get_global_place_group(
                place_group_id, base_url, load_features=load_features
            )

        place_group_id = self.get_place_group_id_safe(place_group_config)

        place_group = self.get_cached_place_group(place_group_id)
        if place_group is None:
            place_group_title = place_group_config.get("Title", place_group_id)
            place_group_description = place_group_config.get("Description")
            place_path_wc = self.get_config_path(
                place_group_config, "'PlaceGroups' item"
            )
            fs, place_path = fsspec.core.url_to_fs(place_path_wc)
            source_paths = [fs.unstrip_protocol(p) for p in fs.glob(place_path)]
            source_encoding = place_group_config.get("CharacterEncoding", "utf-8")

            join = None
            place_join = place_group_config.get("Join")
            if isinstance(place_join, dict):
                join_path = self.get_config_path(
                    place_join, "'Join' of a 'PlaceGroups' item"
                )
                join_property = place_join.get("Property")
                if not join_property:
                    raise ApiError.InvalidServerConfig(
                        "Missing 'Property' entry in 'Join' of a 'PlaceGroups' item"
                    )
                join_encoding = place_join.get("CharacterEncoding", "utf-8")
                join = dict(
                    path=join_path, property=join_property, encoding=join_encoding
                )

            property_mapping = self.get_property_mapping(base_url, place_group_config)

            place_group = dict(
                type="FeatureCollection",
                features=None,
                id=place_group_id,
                title=place_group_title,
                description=place_group_description,
                propertyMapping=property_mapping,
                sourcePaths=source_paths,
                sourceEncoding=source_encoding,
                join=join,
            )

            self.check_sub_group_configs(place_group_config)
            self.set_cached_place_group(place_group_id, place_group)

        if load_features:
            self.load_place_group_features(place_group)

        return place_group

    @staticmethod
    def get_place_group_id_safe(place_group_config):
        place_group_id = place_group_config.get("Identifier")
        if not place_group_id:
            raise ApiError.InvalidServerConfig(
                "Missing 'Identifier' entry in a 'PlaceGroups' item"
            )
        return place_group_id

    @staticmethod
    def check_sub_group_configs(place_group_config):
        sub_place_group_configs = place_group_config.get("Places")
        if sub_place_group_configs:
            raise ApiError.InvalidServerConfig(
                "Invalid 'Places' entry in a 'PlaceGroups' item: not implemented yet"
            )

    @staticmethod
    def get_property_mapping(base_url, place_group_config):
        property_mapping = place_group_config.get("PropertyMapping")
        if property_mapping:
            property_mapping = dict(property_mapping)
            for key, value in property_mapping.items():
                if isinstance(value, str) and "${base_url}" in value:
                    property_mapping[key] = value.replace("${base_url}", base_url)
        return property_mapping

    def load_place_group_features(
        self, place_group: PlaceGroup
    ) -> list[dict[str, Any]]:
        features = place_group.get("features")
        if features is not None:
            return features
        source_paths = place_group["sourcePaths"]
        source_encoding = place_group["sourceEncoding"]
        features = []
        for source_path in source_paths:
            with self.measure_time(f"Loading features from file {source_path}"):
                with fiona.open(
                    source_path, encoding=source_encoding
                ) as feature_collection:
                    for feature in self._to_geo_interface(feature_collection):
                        self._remove_feature_id(feature)
                        feature["id"] = self.new_feature_id()
                        features.append(feature)

        join = place_group["join"]
        if join:
            join_path = join["path"]
            join_property = join["property"]
            join_encoding = join["encoding"]
            with fiona.open(join_path, encoding=join_encoding) as feature_collection:
                indexed_join_features = self._get_indexed_features(
                    list(self._to_geo_interface(feature_collection)), join_property
                )
            for feature in features:
                properties = feature.get("properties")
                if isinstance(properties, dict) and join_property in properties:
                    join_value = properties[join_property]
                    join_feature = indexed_join_features.get(join_value)
                    if join_feature:
                        join_properties = join_feature.get("properties")
                        if join_properties:
                            properties.update(join_properties)
                            feature["properties"] = properties

        place_group["features"] = features
        return features

    @classmethod
    def _to_geo_interface(
        cls, feature_collection: fiona.Collection
    ) -> Iterator[dict[str, Any]]:
        source_crs = feature_collection.crs
        target_crs = fiona.crs.CRS.from_epsg(4326)
        project: Callable | None = None
        if not source_crs == target_crs:
            project = pyproj.Transformer.from_crs(
                source_crs, target_crs, always_xy=True
            ).transform
        for feature in feature_collection:
            # fiona >=1.9 returns features of type fiona.model.Feature
            # rather than JSON-serializable dictionaries.
            if hasattr(feature, "__geo_interface__"):
                # We fall back on the traditional geo-interface:
                feature = feature.__geo_interface__
                # Fiona =1.9.0 adds empty "geometries" field
                # to any "geometry", we fix this too:
                geometry = feature.get("geometry")
                if (
                    geometry
                    and "geometries" in geometry
                    and geometry.get("type") != "GeometryCollection"
                ):
                    del geometry["geometries"]
            if project is not None:
                geometry = feature.get("geometry")
                shapely_geom = shapely.geometry.shape(geometry)
                feature["geometry"] = shapely.ops.transform(
                    project, shapely_geom
                ).__geo_interface__
            yield feature

    @classmethod
    def _get_indexed_features(
        cls, features: Sequence[Feature], property_name: str
    ) -> Feature:
        feature_index = {}
        for feature in features:
            properties = feature.get("properties")
            if properties and property_name in properties:
                property_value = properties[property_name]
                feature_index[property_value] = feature
        return feature_index

    @classmethod
    def _remove_feature_id(cls, feature: Feature):
        if "id" in feature:
            del feature["id"]
        if "ID" in feature:
            del feature["ID"]



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from .cached import CachedZarrStore
from .diagnostic import DiagnosticZarrStore
from .generic import GenericArray, GenericArrayLike, GenericZarrStore
from .holder import ZarrStoreHolder
from .logging import LoggingZarrStore



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from xcube.util.jsonschema import (
    JsonArraySchema,
    JsonComplexSchema,
    JsonNumberSchema,
    JsonObjectSchema,
)
from xcube.webapi.common.schemas import (
    BOOLEAN_SCHEMA,
    CHUNK_SIZE_SCHEMA,
    FILE_SYSTEM_SCHEMA,
    GEO_BOUNDING_BOX_SCHEMA,
    IDENTIFIER_SCHEMA,
    PATH_SCHEMA,
    STRING_SCHEMA,
    URI_SCHEMA,
    NUMBER_SCHEMA,
)

from ..places.config import PLACE_GROUP_SCHEMA

ATTRIBUTION_SCHEMA = JsonComplexSchema(
    one_of=[
        STRING_SCHEMA,
        JsonArraySchema(items=STRING_SCHEMA),
    ]
)

VARIABLES_SCHEMA = JsonArraySchema(
    items=IDENTIFIER_SCHEMA,
    min_items=1,
    description="Names of variables to be published."
    " Names may use wildcard characters '*' and '?'."
    " Also determines the order of variables.",
)

VALUE_RANGE_SCHEMA = JsonArraySchema(items=[JsonNumberSchema(), JsonNumberSchema()])

AUGMENTATION_SCHEMA = JsonObjectSchema(
    properties=dict(
        Path=PATH_SCHEMA,
        Function=IDENTIFIER_SCHEMA,
        Class=IDENTIFIER_SCHEMA,
        InputParameters=JsonObjectSchema(
            additional_properties=True,
        ),
    ),
    required=[
        "Path",
        "Function",
    ],
    additional_properties=False,
)

ACCESS_CONTROL_SCHEMA = JsonObjectSchema(
    properties=dict(
        IsSubstitute=BOOLEAN_SCHEMA,
        RequiredScopes=JsonArraySchema(items=IDENTIFIER_SCHEMA),
    ),
    additional_properties=False,
)

COMMON_DATASET_PROPERTIES = dict(
    Title=STRING_SCHEMA,
    Description=STRING_SCHEMA,
    GroupTitle=STRING_SCHEMA,
    GroupId=STRING_SCHEMA,
    SortValue=NUMBER_SCHEMA,
    Tags=JsonArraySchema(items=STRING_SCHEMA),
    Variables=VARIABLES_SCHEMA,
    TimeSeriesDataset=IDENTIFIER_SCHEMA,
    BoundingBox=GEO_BOUNDING_BOX_SCHEMA,
    ChunkCacheSize=CHUNK_SIZE_SCHEMA,
    Augmentation=AUGMENTATION_SCHEMA,
    Style=IDENTIFIER_SCHEMA,
    Hidden=BOOLEAN_SCHEMA,
    AccessControl=ACCESS_CONTROL_SCHEMA,
    PlaceGroups=JsonArraySchema(items=PLACE_GROUP_SCHEMA),
    Attribution=ATTRIBUTION_SCHEMA,
)

DATASET_CONFIG_SCHEMA = JsonObjectSchema(
    properties=dict(
        Identifier=IDENTIFIER_SCHEMA,
        StoreInstanceId=IDENTIFIER_SCHEMA,  # will be set by server
        Path=PATH_SCHEMA,
        FileSystem=FILE_SYSTEM_SCHEMA,
        Anonymous=BOOLEAN_SCHEMA,
        Endpoint=URI_SCHEMA,
        Region=IDENTIFIER_SCHEMA,
        Function=IDENTIFIER_SCHEMA,
        Class=IDENTIFIER_SCHEMA,
        InputDatasets=JsonArraySchema(items=IDENTIFIER_SCHEMA),
        InputParameters=JsonObjectSchema(additional_properties=True),
        **COMMON_DATASET_PROPERTIES,
    ),
    required=["Identifier", "Path"],
    additional_properties=False,
)

DATA_STORE_DATASET_SCHEMA = JsonObjectSchema(
    required=["Path"],
    properties=dict(
        Identifier=IDENTIFIER_SCHEMA,
        Path=PATH_SCHEMA,
        StoreInstanceId=IDENTIFIER_SCHEMA,  # will be set by server
        StoreOpenParams=JsonObjectSchema(additional_properties=True),
        **COMMON_DATASET_PROPERTIES,
    ),
    additional_properties=False,
)

DATA_STORE_SCHEMA = JsonObjectSchema(
    properties=dict(
        Identifier=IDENTIFIER_SCHEMA,
        StoreId=IDENTIFIER_SCHEMA,
        StoreParams=JsonObjectSchema(additional_properties=True),
        Datasets=JsonArraySchema(items=DATA_STORE_DATASET_SCHEMA),
    ),
    required=[
        "Identifier",
        "StoreId",
    ],
    additional_properties=False,
)

COLOR_MAPPING_EXPLICIT_SCHEMA = JsonObjectSchema(
    properties=dict(ColorBar=STRING_SCHEMA, ValueRange=VALUE_RANGE_SCHEMA),
    required=[],
    additional_properties=False,
)

COLOR_MAPPING_BY_PATH_SCHEMA = JsonObjectSchema(
    properties=dict(
        ColorFile=STRING_SCHEMA,
    ),
    required=[
        "ColorFile",
    ],
    additional_properties=False,
)

COLOR_MAPPING_SCHEMA = JsonComplexSchema(
    one_of=[
        COLOR_MAPPING_EXPLICIT_SCHEMA,
        COLOR_MAPPING_BY_PATH_SCHEMA,
    ]
)

CHANNEL_MAPPING_SCHEMA = JsonObjectSchema(
    properties=dict(
        ValueRange=JsonArraySchema(items=[JsonNumberSchema(), JsonNumberSchema()]),
        Variable=STRING_SCHEMA,
    ),
    required=["ValueRange", "Variable"],
    additional_properties=False,
)

RGB_MAPPING_SCHEMA = JsonObjectSchema(
    properties=dict(
        Red=CHANNEL_MAPPING_SCHEMA,
        Green=CHANNEL_MAPPING_SCHEMA,
        Blue=CHANNEL_MAPPING_SCHEMA,
    ),
    required=[
        "Red",
        "Green",
        "Blue",
    ],
    additional_properties=False,
)

STYLE_SCHEMA = JsonObjectSchema(
    properties=dict(
        Identifier=STRING_SCHEMA,
        ColorMappings=JsonObjectSchema(
            properties=dict(rgb=RGB_MAPPING_SCHEMA),
            additional_properties=COLOR_MAPPING_SCHEMA,
        ),
    ),
    required=[
        "Identifier",
        "ColorMappings",
    ],
    additional_properties=False,
)

COLOR_SCHEMA = JsonComplexSchema(
    one_of=[
        STRING_SCHEMA,
        JsonArraySchema(
            items=JsonNumberSchema(minimum=0, maximum=255), min_items=3, max_items=4
        ),
    ]
)

CUSTOM_COLOR_ENTRY_SCHEMA = JsonObjectSchema(
    properties=dict(
        Value=JsonNumberSchema(),
        Color=COLOR_SCHEMA,
        Label=STRING_SCHEMA,
    ),
    required=["Value", "Color"],
    additional_properties=False,
)

CUSTOM_COLOR_LIST_SCHEMA = JsonComplexSchema(
    one_of=[
        JsonArraySchema(
            items=[JsonNumberSchema(), COLOR_SCHEMA], min_items=2, max_items=2
        ),
        JsonArraySchema(
            items=[JsonNumberSchema(), COLOR_SCHEMA, STRING_SCHEMA],
            min_items=3,
            max_items=3,
        ),
    ]
)

CUSTOM_COLORS_SCHEMA = JsonComplexSchema(
    one_of=[
        CUSTOM_COLOR_ENTRY_SCHEMA,
        CUSTOM_COLOR_LIST_SCHEMA,
    ]
)

CUSTOM_COLORMAP_SCHEMA = JsonObjectSchema(
    properties=dict(
        Identifier=STRING_SCHEMA,
        Type=STRING_SCHEMA,
        Colors=JsonArraySchema(items=CUSTOM_COLORS_SCHEMA, min_items=1),
    ),
    required=["Identifier", "Type", "Colors"],
    additional_properties=False,
)

SERVICE_PROVIDER_SCHEMA = JsonObjectSchema(
    additional_properties=True,
)

ENTRYPOINT_DATASET_ID_SCHEMA = JsonObjectSchema(
    properties=dict(
        Identifier=IDENTIFIER_SCHEMA,
    ),
    required=["Identifier"],
    additional_properties=False,
)

DATASET_GROUPS_SCHEMA = JsonObjectSchema(
    properties=dict(
        Identifier=IDENTIFIER_SCHEMA, Title=STRING_SCHEMA, Description=STRING_SCHEMA
    ),
    required=["Identifier", "Title"],
    additional_properties=False,
)

CONFIG_SCHEMA = JsonObjectSchema(
    properties=dict(
        DatasetAttribution=ATTRIBUTION_SCHEMA,
        AccessControl=ACCESS_CONTROL_SCHEMA,
        DatasetChunkCacheSize=CHUNK_SIZE_SCHEMA,
        DatasetGroups=JsonArraySchema(items=DATASET_GROUPS_SCHEMA),
        Datasets=JsonArraySchema(items=DATASET_CONFIG_SCHEMA),
        DataStores=JsonArraySchema(items=DATA_STORE_SCHEMA),
        Styles=JsonArraySchema(items=STYLE_SCHEMA),
        CustomColorMaps=JsonArraySchema(items=CUSTOM_COLORMAP_SCHEMA),
        ServiceProvider=SERVICE_PROVIDER_SCHEMA,
        EntrypointDatasetId=IDENTIFIER_SCHEMA,
    )
)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import sys
from collections.abc import Mapping, MutableMapping
from contextlib import contextmanager
from functools import cached_property
from pathlib import Path
from typing import Any, Optional

import fsspec
from chartlets import Extension, ExtensionContext

from xcube.constants import LOG
from xcube.server.api import Context
from xcube.util.fspath import is_local_fs
from xcube.util.temp import new_temp_dir
from xcube.webapi.common.context import ResourcesContext
from xcube.webapi.viewer.contrib import Panel

Extension.add_contrib_point("panels", Panel)


class ViewerContext(ResourcesContext):
    """Context for xcube Viewer API."""

    def __init__(self, server_ctx: Context):
        super().__init__(server_ctx)
        self.ext_ctx: ExtensionContext | None = None
        self.persistence: MutableMapping | None = None

    def on_update(self, prev_context: Optional[Context]):
        super().on_update(prev_context)
        viewer_config: dict = self.config.get("Viewer")
        if not viewer_config:
            return
        persistence: dict | None = viewer_config.get("Persistence")
        if persistence:
            path = self.get_config_path(persistence, "Persistence")
            storage_options = persistence.get("StorageOptions")
            self.set_persistence(path, storage_options)
        augmentation: dict | None = viewer_config.get("Augmentation")
        if augmentation:
            path = augmentation.get("Path")
            extension_refs = augmentation["Extensions"]
            self.set_extension_context(path, extension_refs)

    @cached_property
    def config_items(self) -> Optional[Mapping[str, bytes]]:
        if self.config_path is None:
            return None
        return fsspec.get_mapper(self.config_path)

    @cached_property
    def config_path(self) -> Optional[str]:
        if "Viewer" not in self.config:
            return None
        return self.get_config_path(
            self.config["Viewer"].get("Configuration", {}),
            "'Configuration' item of 'Viewer'",
        )

    def set_persistence(self, path: str, storage_options: dict[str, Any] | None):
        fs_root: tuple[fsspec.AbstractFileSystem, str] = fsspec.core.url_to_fs(
            path, **(storage_options or {})
        )
        fs, root = fs_root
        self.persistence = fs.get_mapper(root, create=True, check=True)
        LOG.info(f"Viewer persistence established for path {path!r}")

    def set_extension_context(self, path: str | None, extension_refs: list[str]):
        module_path = self.base_dir
        if path:
            module_path = f"{module_path}/{path}"
        fs_root: tuple[fsspec.AbstractFileSystem, str] = fsspec.core.url_to_fs(
            module_path
        )
        fs, fs_path = fs_root
        if is_local_fs(fs):
            local_module_path = Path(module_path)
        else:
            temp_module_path = new_temp_dir("xcube-viewer-aux-")
            LOG.warning(f"Downloading {module_path!r} to {temp_module_path!r}")
            fs.get(fs_path + "/**/*", temp_module_path + "/", recursive=True)
            local_module_path = Path(temp_module_path)
        with prepend_sys_path(local_module_path):
            LOG.info(f"Loading viewer extension(s) {','.join(extension_refs)}")
            self.ext_ctx = ExtensionContext.load(self.server_ctx, extension_refs)


@contextmanager
def prepend_sys_path(path: Path | str | None):
    prev_sys_path = None
    if path is not None:
        LOG.warning(f"Temporarily prepending '{path}' to sys.path")
        prev_sys_path = sys.path
        sys.path = [str(path)] + sys.path
    try:
        yield path is not None
    finally:
        if prev_sys_path:
            sys.path = prev_sys_path
            LOG.info("Restored sys.path")



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from typing import Optional, Tuple, Type, TypeVar, Union

from xcube.util.assertions import assert_true

T = TypeVar("T")

ItemType = Union[type[T], tuple[type[T], ...]]
Pair = tuple[T, T]
ScalarOrPair = Union[T, Pair]


def normalize_scalar_or_pair(
    value: ScalarOrPair[T],
    *,
    item_type: Optional[ItemType[T]] = None,
    name: Optional[str] = None,
) -> Pair:
    try:
        assert_true(
            len(value) <= 2,
            message=f"{name or 'Value'} must be a scalar or pair of "
            f"{item_type or 'scalars'}, was '{value}'",
        )
        x, y = value
    except TypeError:
        x, y = value, value
    if item_type is not None:
        assert_true(
            isinstance(x, item_type) and isinstance(y, item_type),
            message=f"{name or 'Value'} must be a scalar or pair of "
            f"{item_type}, was '{value}'",
        )
    return x, y



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from typing import Tuple, Union

import dask.array as da
import pyproj
import xarray as xr

from xcube.util.assertions import assert_true

from .base import GridMapping
from .helpers import (
    _default_xy_dim_names,
    _default_xy_var_names,
    _normalize_crs,
    _normalize_int_pair,
    _normalize_number_pair,
    _to_int_or_float,
)


class RegularGridMapping(GridMapping):
    def __init__(self, **kwargs):
        kwargs.pop("is_regular", None)
        super().__init__(is_regular=True, **kwargs)
        self._xy_coords = None

    def _new_x_coords(self) -> xr.DataArray:
        self._assert_regular()
        x_res = self.x_res
        x1, x2 = self.x_min + x_res / 2, self.x_max - x_res / 2
        x_name, _ = self.xy_dim_names
        return xr.DataArray(
            da.linspace(x1, x2, self.width, chunks=self.tile_width),
            dims=self.xy_dim_names[0],
        )

    def _new_y_coords(self) -> xr.DataArray:
        self._assert_regular()
        y_res = self.y_res
        y1, y2 = self.y_min + y_res / 2, self.y_max - y_res / 2
        if not self.is_j_axis_up:
            y1, y2 = y2, y1
        return xr.DataArray(
            da.linspace(y1, y2, self.height, chunks=self.tile_height),
            dims=self.xy_dim_names[1],
        )

    def _new_xy_coords(self) -> xr.DataArray:
        self._assert_regular()
        x_coords_1d = da.asarray(self.x_coords.data).rechunk(self.tile_width)
        y_coords_1d = da.expand_dims(
            da.asarray(self.y_coords.data).rechunk(self.tile_height), 1
        )
        y_coords_2d, x_coords_2d = da.broadcast_arrays(y_coords_1d, x_coords_1d)
        xy_coords = da.concatenate(
            [da.expand_dims(x_coords_2d, 0), da.expand_dims(y_coords_2d, 0)]
        )
        xy_coords = da.rechunk(
            xy_coords, chunks=(2, xy_coords.chunksize[1], xy_coords.chunksize[2])
        )
        xy_coords = xr.DataArray(
            xy_coords,
            dims=("coord", self.y_coords.dims[0], self.x_coords.dims[0]),
            name="xy_coords",
        )
        xy_coords.name = "xy_coords"
        return xy_coords


def new_regular_grid_mapping(
    size: Union[int, tuple[int, int]],
    xy_min: tuple[float, float],
    xy_res: Union[float, tuple[float, float]],
    crs: Union[str, pyproj.crs.CRS],
    *,
    tile_size: Union[int, tuple[int, int]] = None,
    is_j_axis_up: bool = False,
) -> GridMapping:
    width, height = _normalize_int_pair(size, name="size")
    assert_true(width > 1 and height > 1, "invalid size")

    x_min, y_min = _normalize_number_pair(xy_min, name="xy_min")

    x_res, y_res = _normalize_number_pair(xy_res, name="xy_res")
    assert_true(x_res > 0 and y_res > 0, "invalid xy_res")

    crs = _normalize_crs(crs)

    x_min = _to_int_or_float(x_min)
    y_min = _to_int_or_float(y_min)
    x_max = _to_int_or_float(x_min + x_res * width)
    y_max = _to_int_or_float(y_min + y_res * height)

    if crs.is_geographic:
        # TODO: don't do that.
        #  Instead set NaN in coord vars returned by to_coords()
        if y_min < -90:
            raise ValueError("invalid y_min")
        if y_max > 90:
            raise ValueError("invalid size, y_min combination")

    return RegularGridMapping(
        crs=crs,
        size=(width, height),
        tile_size=tile_size or (width, height),
        xy_bbox=(x_min, y_min, x_max, y_max),
        xy_res=(x_res, y_res),
        xy_var_names=_default_xy_var_names(crs),
        xy_dim_names=_default_xy_dim_names(crs),
        is_lon_360=(x_max > 180) and crs.is_geographic,
        is_j_axis_up=is_j_axis_up,
    )


def to_regular_grid_mapping(
    grid_mapping: GridMapping,
    *,
    tile_size: Union[int, tuple[int, int]] = None,
    is_j_axis_up: bool = False,
) -> GridMapping:
    if grid_mapping.is_regular:
        if tile_size is not None or is_j_axis_up != grid_mapping.is_j_axis_up:
            return grid_mapping.derive(tile_size=tile_size, is_j_axis_up=is_j_axis_up)
        return grid_mapping

    x_min, y_min, x_max, y_max = grid_mapping.xy_bbox
    x_res, y_res = grid_mapping.xy_res
    # x_digits = 2 + abs(round(math.log10(x_res)))
    # y_digits = 2 + abs(round(math.log10(y_res)))
    # x_min = floor_to_fraction(x_min, x_digits)
    # y_min = floor_to_fraction(y_min, y_digits)
    # x_max = ceil_to_fraction(x_max, x_digits)
    # y_max = ceil_to_fraction(y_max, y_digits)
    xy_res = min(x_res, y_res) or max(x_res, y_res)
    width = round((x_max - x_min + xy_res) / xy_res)
    height = round((y_max - y_min + xy_res) / xy_res)
    width = width if width >= 2 else 2
    height = height if height >= 2 else 2

    if tile_size is None:
        tile_size = grid_mapping.tile_size
    return new_regular_grid_mapping(
        size=(width, height),
        xy_min=(x_min, y_min),
        xy_res=xy_res,
        crs=grid_mapping.crs,
        tile_size=tile_size,
        is_j_axis_up=is_j_axis_up,
    )



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import json
import os.path
import sys
from collections.abc import Sequence
from typing import Any, Dict, Optional, Union

import jsonschema
import yaml

from xcube.core.byoa import CodeConfig
from xcube.util.assertions import assert_false, assert_instance, assert_true
from xcube.util.jsonschema import JsonArraySchema, JsonObject, JsonObjectSchema

from ...constants import LOG
from .config import CallbackConfig, CubeConfig, InputConfig, OutputConfig
from .error import CubeGeneratorError

CubeGeneratorRequestLike = Union[str, dict, "CubeGeneratorRequest"]


class CubeGeneratorRequest(JsonObject):
    """
    A request used to generate data cubes using cube generators.

    Args:
        input_config: A configuration for a single input.
            Must be omitted if *input_configs* is given.
        input_configs: A sequence of one or more input configurations.
            Must be omitted if *input_config* is given.
        cube_config: The target cube configuration.
        code_config: The user-code configuration.
        output_config: The output configuration for the target cube.
        callback_config: A configuration that allows a cube generator
            to publish progress information to a compatible endpoint.
    """

    def __init__(
        self,
        input_config: InputConfig = None,
        input_configs: Sequence[InputConfig] = None,
        cube_config: CubeConfig = None,
        code_config: CodeConfig = None,
        output_config: OutputConfig = None,
        callback_config: Optional[CallbackConfig] = None,
    ):
        assert_true(
            input_config or input_configs,
            "one of input_config and input_configs must be given",
        )
        assert_false(
            input_config and input_configs,
            "input_config and input_configs cannot be given both",
        )
        if input_config is not None:
            assert_instance(input_config, InputConfig, "input_config")
            input_configs = [input_config]
        elif input_configs is not None:
            assert_instance(input_configs, (list, tuple), "input_configs")
            for i in range(len(input_configs)):
                assert_instance(input_configs[i], InputConfig, f"input_configs[{i}]")
        if cube_config is not None:
            assert_instance(cube_config, CubeConfig, "cube_config")
        if code_config is not None:
            assert_instance(code_config, CodeConfig, "code_config")
        assert_instance(output_config, OutputConfig, "output_config")
        if callback_config is not None:
            assert_instance(callback_config, CallbackConfig, "callback_config")
        self.input_configs = input_configs
        self.cube_config = cube_config
        self.code_config = code_config
        self.output_config = output_config
        self.callback_config = callback_config

    def for_service(self) -> "CubeGeneratorRequest":
        if self.code_config is None:
            return self
        return CubeGeneratorRequest(
            input_configs=self.input_configs,
            cube_config=self.cube_config,
            code_config=self.code_config.for_service(),
            output_config=self.output_config,
            callback_config=self.callback_config,
        )

    def for_local(self) -> "CubeGeneratorRequest":
        if self.code_config is None:
            return self
        return CubeGeneratorRequest(
            input_configs=self.input_configs,
            cube_config=self.cube_config,
            code_config=self.code_config.for_local(),
            output_config=self.output_config,
            callback_config=self.callback_config,
        )

    @classmethod
    def get_schema(cls):
        return JsonObjectSchema(
            properties=dict(
                input_config=InputConfig.get_schema(),
                input_configs=JsonArraySchema(
                    items=InputConfig.get_schema(), min_items=1
                ),
                cube_config=CubeConfig.get_schema(),
                code_config=CodeConfig.get_schema(),
                output_config=OutputConfig.get_schema(),
                callback_config=CallbackConfig.get_schema(),
            ),
            required=["output_config"],
            factory=cls,
        )

    @classmethod
    def normalize(cls, request: CubeGeneratorRequestLike) -> "CubeGeneratorRequest":
        """Normalize given *request* to an instance of
        :class:`CubeGeneratorRequest`.

        If *request* is already a CubeGeneratorRequest it is returned as is.
        If it is a ``str``, it is interpreted as a YAML or JSON file path
        and the request is read from file using
        ``CubeGeneratorRequest.from_file()``.
        If it is a ``dict``, it is interpreted as a JSON object and the
        request is parsed using ``CubeGeneratorRequest.from_dict()``.

        Args:
            request: The request, or request file path,
                or request JSON object.
        Raises:
            TypeError: if *request* is not a ``CubeGeneratorRequest``,
                ``str``, or ``dict``.
        """
        if isinstance(request, CubeGeneratorRequest):
            return request
        if isinstance(request, str):
            return CubeGeneratorRequest.from_file(request)
        if isinstance(request, dict):
            return CubeGeneratorRequest.from_dict(request)
        raise TypeError(
            "request must be a str, dict, or a CubeGeneratorRequest instance"
        )

    def to_dict(self) -> dict[str, Any]:
        """Convert into a JSON-serializable dictionary"""
        d = {}

        if self.input_configs is not None:
            if len(self.input_configs) == 1:
                d.update(input_config=self.input_configs[0].to_dict())
            else:
                d.update(input_configs=[ic.to_dict() for ic in self.input_configs])

        if self.cube_config is not None:
            d.update(cube_config=self.cube_config.to_dict())

        if self.code_config is not None:
            d.update(code_config=self.code_config.to_dict())

        d.update(output_config=self.output_config.to_dict())

        if self.callback_config is not None:
            d.update(callback_config=self.callback_config.to_dict())

        return d

    @classmethod
    def from_dict(cls, request_dict: dict[str, Any]) -> "CubeGeneratorRequest":
        """Create new instance from a JSON-serializable dictionary"""
        try:
            return cls.get_schema().from_instance(request_dict)
        except jsonschema.exceptions.ValidationError as e:
            raise CubeGeneratorError(f"{e}", status_code=400) from e

    @classmethod
    def from_file(
        cls, request_file: Optional[str], verbosity: int = 0
    ) -> "CubeGeneratorRequest":
        """
        Create new instance from a JSON file, or YAML file,
        or JSON passed via stdin.
        """
        request_dict = cls._load_request_file(request_file, verbosity=verbosity)
        if verbosity:
            LOG.info(f"Cube generator request loaded from {request_file or 'TTY'}.")
        return cls.from_dict(request_dict)

    @classmethod
    def _load_request_file(
        cls, gen_config_file: Optional[str], verbosity: int = 0
    ) -> dict:
        if gen_config_file is not None and not os.path.exists(gen_config_file):
            raise CubeGeneratorError(
                f'Cube generator request "{gen_config_file}" not found.'
            )

        try:
            if gen_config_file is None:
                if not sys.stdin.isatty():
                    if verbosity:
                        LOG.info("Awaiting generator request JSON from TTY...")
                    return json.load(sys.stdin)
            else:
                with open(gen_config_file) as fp:
                    if gen_config_file.endswith(".json"):
                        return json.load(fp)
                    else:
                        return yaml.safe_load(fp)
        except BaseException as e:
            raise CubeGeneratorError(
                f'Error loading generator request "{gen_config_file}": {e}'
            ) from e

        raise CubeGeneratorError(f"Missing cube generator request.")



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import os.path
from typing import Optional

import click

YAML_FORMAT = "yaml"
JSON_FORMAT = "json"
DEFAULT_FORMAT = YAML_FORMAT


# noinspection PyShadowingBuiltins
@click.command(name="versions")
@click.option(
    "--format",
    "-f",
    "format_name",
    metavar="FORMAT",
    type=click.Choice([YAML_FORMAT, JSON_FORMAT], case_sensitive=False),
    help=f"Output format. "
    f"Must be one of {YAML_FORMAT!r} or {JSON_FORMAT!r}. "
    f"If not given, derived from OUTPUT name extension.",
)
@click.option(
    "--output", "-o", "output_path", metavar="OUTPUT", help=f"Output file path."
)
def versions(format_name: Optional[str], output_path: Optional[str]):
    """
    Get versions of important packages used by xcube.
    """
    from xcube.util.versions import get_xcube_versions

    xcube_versions = get_xcube_versions()

    if not format_name:
        format_name = YAML_FORMAT
        if output_path:
            _, ext = os.path.splitext(output_path)
            if ext.lower() == ".json":
                format_name = JSON_FORMAT

    if format_name.lower() == JSON_FORMAT:
        import json

        text = json.dumps(xcube_versions, indent=2)
    else:
        import yaml

        text = yaml.dump(xcube_versions)

    if output_path:
        with open(output_path, "w") as fp:
            fp.write(text)
    else:
        print(text)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import collections.abc
import concurrent.futures
import copy
from collections.abc import Awaitable, Mapping, Sequence
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

import jsonschema.exceptions

from xcube.constants import EXTENSION_POINT_SERVER_APIS, LOG
from xcube.util.assertions import assert_instance, assert_subclass
from xcube.util.extension import ExtensionRegistry
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.plugin import get_extension_registry
from xcube.version import version

from ..util.frozen import FrozenDict
from .api import (
    Api,
    ApiContext,
    ApiContextT,
    ApiRoute,
    ApiStaticRoute,
    Context,
    ReturnT,
    ServerConfig,
)
from .asyncexec import AsyncExecution
from .config import BASE_SERVER_CONFIG_SCHEMA, get_url_prefix, resolve_config_path
from .framework import Framework


class Server(AsyncExecution):
    """A REST server extendable by API extensions.

    APIs are registered using the extension point
    "xcube.server.api".

    TODO:
      * Allow server to serve generic static content,
        e.g. "http://localhost:8080/images/outside-cube/${ID}.jpg"
      * Allow server updates triggered by local file changes
        and changes in s3 buckets
      * Address common server configuration
        - Configure server types by meta-configuration,
          e.g. server name, server description, names of APIs served,
          aliases for common server config...
        - Why we need aliases for common server config:
          o Camel case vs snake case parameters names,
            e.g. "BaseDir" vs "base_dir"
          o First capital letter in parameter names,
            e.g. "Address" vs "address"
      * Use any given request JSON schema in openAPI
        to validate requests in HTTP methods

    Args:
        framework: The web server framework to be used
        config: The server configuration.
        extension_registry: Optional extension registry. Defaults to
            xcube's default extension registry.
    """

    def __init__(
        self,
        framework: Framework,
        config: Mapping[str, Any],
        extension_registry: Optional[ExtensionRegistry] = None,
    ):
        assert_instance(framework, Framework)
        assert_instance(config, collections.abc.Mapping)
        apis = self.load_apis(config, extension_registry=extension_registry)
        for api in apis:
            LOG.info(f"Loaded service API {api.name!r}")
        static_routes = self._collect_static_routes(config)
        static_routes.extend(self._collect_api_static_routes(apis))
        routes = self._collect_api_routes(apis)
        url_prefix = get_url_prefix(config)
        framework.add_routes(routes, url_prefix)
        framework.add_static_routes(static_routes, url_prefix)
        self._framework = framework
        self._apis = apis
        self._config_schema = self.get_effective_config_schema(framework, apis)
        ctx = self._new_ctx(config)
        ctx.on_update(None)
        self._set_ctx(ctx)

    @property
    def framework(self) -> Framework:
        """The web server framework used by this server."""
        return self._framework

    @property
    def apis(self) -> tuple[Api]:
        """The APIs supported by this server."""
        return self._apis

    @property
    def config_schema(self) -> JsonObjectSchema:
        """The effective JSON schema for the server configuration."""
        return self._config_schema

    @property
    def ctx(self) -> "ServerContext":
        """The current server context."""
        return self._ctx

    def _set_ctx(self, ctx: "ServerContext"):
        self._ctx = ctx
        self._framework.update(ctx)

    def _new_ctx(self, config: collections.abc.Mapping) -> "ServerContext":
        config = dict(config)
        for key in tuple(config.keys()):
            if key not in self._config_schema.properties:
                LOG.warning(
                    f"Configuration setting {key!r} ignored,"
                    f" because there is no schema describing it."
                )
                config.pop(key)
        try:
            validated_config = self._config_schema.from_instance(config)
        except jsonschema.exceptions.ValidationError as e:
            raise ValueError(f"Invalid server configuration:\n{e}") from e
        return ServerContext(self, validated_config)

    def start(self):
        """Start this server."""
        LOG.info(f"Starting service...")
        for api in self._apis:
            api.on_start(self.ctx)
        self._framework.start(self.ctx)

    def stop(self):
        """Stop this server."""
        LOG.info(f"Stopping service...")
        self._framework.stop(self.ctx)
        for api in self._apis:
            api.on_stop(self.ctx)
        self._ctx.on_dispose()

    def update(self, config: Mapping[str, Any]):
        """Update this server with given server configuration."""
        ctx = self._new_ctx(config)
        ctx.on_update(prev_ctx=self._ctx)
        self._set_ctx(ctx)

    def call_later(self, delay: Union[int, float], callback: Callable, *args, **kwargs):
        """Executes the given callable *callback* after *delay* seconds.

        Args:
            delay: Delay in seconds.
            callback: Callback to be called.
            *args: Positional arguments passed to *callback*.
            **kwargs: Keyword arguments passed to *callback*.
        """
        return self._framework.call_later(delay, callback, *args, **kwargs)

    def run_in_executor(
        self,
        executor: Optional[concurrent.futures.Executor],
        function: Callable[..., ReturnT],
        *args: Any,
        **kwargs: Any,
    ) -> Awaitable[ReturnT]:
        """Concurrently runs a *function* in a ``concurrent.futures.Executor``.
        If *executor* is ``None``, the framework's default
        executor will be used.

        Args:
            executor: An optional executor.
            function: The function to be run concurrently.
            *args: Positional arguments passed to *function*.
            **kwargs: Keyword arguments passed to *function*.

        Returns:
            The awaitable return value of *function*.
        """
        return self._framework.run_in_executor(executor, function, *args, **kwargs)

    @classmethod
    def load_apis(
        cls,
        config: collections.abc.Mapping,
        extension_registry: Optional[ExtensionRegistry] = None,
    ) -> tuple[Api]:
        # Collect all registered API extensions
        extension_registry = extension_registry or get_extension_registry()
        api_extensions = extension_registry.find_extensions(EXTENSION_POINT_SERVER_APIS)

        # Get APIs specification
        api_spec = config.get("api_spec", {})
        incl_api_names = api_spec.get("includes", [ext.name for ext in api_extensions])
        excl_api_names = api_spec.get("excludes", [])

        # Collect effective APIs
        api_names = set(incl_api_names).difference(set(excl_api_names))
        apis: list[Api] = [
            ext.component for ext in api_extensions if ext.name in api_names
        ]

        api_lookup = {api.name: api for api in apis}

        def assert_required_apis_available():
            # Assert that required APIs are available.
            for api in apis:
                for req_api_name in api.required_apis:
                    if req_api_name not in api_lookup:
                        raise ValueError(
                            f"API {api.name!r}: missing API dependency {req_api_name!r}"
                        )

        assert_required_apis_available()

        def count_api_refs(api: Api) -> int:
            # Count the number of times the given API is referenced.
            dep_sum = 0
            for req_api_name in api.required_apis:
                dep_sum += count_api_refs(api_lookup[req_api_name]) + 1
            for opt_api_name in api.optional_apis:
                if opt_api_name in api_lookup:
                    dep_sum += count_api_refs(api_lookup[opt_api_name]) + 1
            return dep_sum

        # Count the number of times each API is referenced.
        api_ref_counts = {api.name: count_api_refs(api) for api in apis}

        # Return an ordered dict sorted by an API's reference count
        return tuple(sorted(apis, key=lambda api: api_ref_counts[api.name]))

    @classmethod
    def _collect_static_routes(
        cls, config: collections.abc.Mapping
    ) -> list[ApiStaticRoute]:
        static_routes = config.get("static_routes", [])
        api_static_routes = []
        for static_route in static_routes:
            params = dict(**static_route)
            dir_path = params.get("dir_path")
            if dir_path is not None:
                dir_path = resolve_config_path(config, dir_path)
                params["dir_path"] = dir_path
            try:
                api_static_route = ApiStaticRoute(**params)
                api_static_routes.append(api_static_route)
            except (TypeError, ValueError):
                LOG.error(f"Failed to add static route: {params!r}", exc_info=True)
        return api_static_routes

    @classmethod
    def _collect_api_static_routes(cls, apis: Sequence[Api]) -> list[ApiStaticRoute]:
        static_routes = []
        for api in apis:
            static_routes.extend(api.static_routes)
        return static_routes

    @classmethod
    def _collect_api_routes(cls, apis: Sequence[Api]) -> list[ApiRoute]:
        handlers = []
        for api in apis:
            handlers.extend(api.routes)
        return handlers

    @classmethod
    def get_effective_config_schema(
        cls, framework: Framework, apis: Sequence[Api]
    ) -> JsonObjectSchema:
        effective_config_schema = copy.deepcopy(BASE_SERVER_CONFIG_SCHEMA)
        framework_config_schema = framework.config_schema
        if framework_config_schema is not None:
            cls._update_config_schema(
                effective_config_schema, framework_config_schema, f"Server"
            )
        for api in apis:
            api_config_schema = api.config_schema
            if api_config_schema is not None:
                cls._update_config_schema(
                    effective_config_schema, api_config_schema, f"API {api.name!r}"
                )
        return effective_config_schema

    @classmethod
    def _update_config_schema(
        cls,
        config_schema: JsonObjectSchema,
        config_schema_update: JsonObjectSchema,
        schema_name: str,
    ):
        assert isinstance(config_schema, JsonObjectSchema)
        assert isinstance(config_schema_update, JsonObjectSchema)
        for k, v in config_schema_update.properties.items():
            if k in config_schema.properties:
                raise ValueError(
                    f"{schema_name}: configuration parameter {k!r} is already defined."
                )
            config_schema.properties[k] = v
        if config_schema_update.required:
            for r in config_schema_update.required:
                if r not in config_schema.required:
                    config_schema.required.append(r)

    def get_open_api_doc(self, include_all: bool = False) -> dict[str, Any]:
        """Get the OpenAPI JSON document for this server."""
        error_schema = {
            "type": "object",
            "properties": {
                "status_code": {
                    "type": "integer",
                    "minimum": 200,
                },
                "message": {
                    "type": "string",
                },
                "reason": {
                    "type": "string",
                },
                "exception": {"type": "array", "items": {"type": "string"}},
            },
            "additionalProperties": True,
            "required": ["status_code", "message"],
        }

        schema_components = {
            "Error": {
                "type": "object",
                "properties": {
                    "error": error_schema,
                },
                "additionalProperties": True,
                "required": ["error"],
            }
        }

        response_components = {
            "UnexpectedError": {
                "description": "Unexpected error.",
                "content": {
                    "application/json": {
                        "schema": {"$ref": "#/components/schemas/Error"}
                    }
                },
            }
        }

        default_responses = {
            "200": {
                "description": "On success.",
            },
            "default": {"$ref": "#/components/responses/UnexpectedError"},
        }

        url_prefix = get_url_prefix(self.ctx.config)

        tags = []
        paths = {}
        for api in self.ctx.apis:
            if not api.routes and not api.static_routes:
                # Only include APIs with endpoints
                continue
            tags.append({"name": api.name, "description": api.description or ""})
            for route in api.routes:
                if not include_all and route.path.startswith("/maintenance/"):
                    continue
                path = dict(description=getattr(route.handler_cls, "__doc__", "") or "")
                for method in ("head", "get", "post", "put", "delete", "options"):
                    fn = getattr(route.handler_cls, method, None)
                    openapi_metadata = getattr(fn, "__openapi__", None)
                    if isinstance(openapi_metadata, dict):
                        openapi_metadata = openapi_metadata.copy()
                        if "tags" not in openapi_metadata:
                            openapi_metadata["tags"] = [api.name]
                        if "description" not in openapi_metadata:
                            openapi_metadata["description"] = (
                                getattr(fn, "__doc__", None) or ""
                            )
                        responses = openapi_metadata.get("responses")
                        if responses is None:
                            responses = default_responses.copy()
                        elif isinstance(responses, dict):
                            _responses = default_responses.copy()
                            _responses.update(responses)
                            responses = _responses
                        openapi_metadata["responses"] = responses
                        path[method] = dict(**openapi_metadata)
                paths[route.path] = path
            for route in api.static_routes:
                openapi_metadata = dict(route.openapi_metadata or {})
                if "tags" not in openapi_metadata:
                    openapi_metadata["tags"] = [api.name]
                paths[route.path] = dict(get=dict(**openapi_metadata))

        return {
            "openapi": "3.0.0",
            "info": {
                "title": "xcube Server",
                "description": "xcube Server API",
                "version": version,
            },
            "servers": [
                {
                    # TODO (forman): the following URL must be adjusted
                    #   e.g. pass request.url_for_path('') as url into
                    #   this method, or even pass the list of servers.
                    "url": f"http://localhost:8080{url_prefix}",
                    "description": "Local development server.",
                },
            ],
            "tags": tags,
            "paths": paths,
            "components": {
                "schemas": schema_components,
                "responses": response_components,
            },
        }


class ServerContext(Context):
    """The server context holds the current server configuration and
    the API context objects that depend on that specific configuration.

    A new server context is created for any new server configuration,
    which in turn will cause all API context objects to be updated.

    The constructor shall not be called directly.

    Args:
        server: The server.
        config: The current server configuration.
    """

    def __init__(self, server: Server, config: collections.abc.Mapping):
        self._server = server
        self._config = FrozenDict.freeze(config)
        self._api_contexts: dict[str, Context] = dict()

    @property
    def server(self) -> Server:
        return self._server

    @property
    def apis(self) -> tuple[Api]:
        return self._server.apis

    def get_open_api_doc(self, include_all: bool = False) -> dict[str, Any]:
        return self._server.get_open_api_doc(include_all=include_all)

    @property
    def config(self) -> ServerConfig:
        return self._config

    def get_api_ctx(
        self, api_name: str, cls: Optional[type[ApiContextT]] = None
    ) -> Optional[ApiContextT]:
        api_ctx = self._api_contexts.get(api_name)
        if cls is not None:
            assert_subclass(cls, ApiContext, name="cls")
            assert_instance(api_ctx, cls, name=f"api_ctx (context of API {api_name!r})")
        return api_ctx

    def _set_api_ctx(self, api_name: str, api_ctx: ApiContext):
        assert_instance(
            api_ctx, ApiContext, name=f"api_ctx (context of API {api_name!r})"
        )
        self._api_contexts[api_name] = api_ctx
        setattr(self, api_name, api_ctx)

    def call_later(
        self, delay: Union[int, float], callback: Callable, *args, **kwargs
    ) -> object:
        return self._server.call_later(delay, callback, *args, **kwargs)

    def run_in_executor(
        self,
        executor: Optional[concurrent.futures.Executor],
        function: Callable[..., ReturnT],
        *args: Any,
        **kwargs: Any,
    ) -> Awaitable[ReturnT]:
        return self._server.run_in_executor(executor, function, *args, **kwargs)

    def on_update(self, prev_ctx: Optional["ServerContext"]):
        if prev_ctx is None:
            LOG.info(f"Applying initial configuration...")
        else:
            LOG.info(f"Applying configuration changes...")
        for api in self.apis:
            prev_api_ctx: Optional[ApiContext] = None
            if prev_ctx is not None:
                prev_api_ctx = prev_ctx.get_api_ctx(api.name)
                assert prev_api_ctx is not None
            for dep_api_name in api.required_apis:
                dep_api_ctx = self.get_api_ctx(dep_api_name)
                assert dep_api_ctx is not None
            next_api_ctx: Optional[ApiContext] = api.create_ctx(self)
            self._set_api_ctx(api.name, next_api_ctx)
            next_api_ctx.on_update(prev_api_ctx)
            if prev_api_ctx is not None and prev_api_ctx is not next_api_ctx:
                prev_api_ctx.on_dispose()

    def on_dispose(self):
        for api_name in reversed([api.name for api in self.apis]):
            api_ctx = self.get_api_ctx(api_name)
            if api_ctx is not None:
                api_ctx.on_dispose()



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import click


# noinspection PyShadowingBuiltins,PyUnusedLocal
@click.command(name="vars2dim")
@click.argument("cube")
@click.option(
    "--variable",
    "--var",
    metavar="VARIABLE",
    default="data",
    help='Name of the new variable that includes all variables. Defaults to "data".',
)
@click.option(
    "--dim_name",
    "-D",
    metavar="DIM-NAME",
    default="var",
    help='Name of the new dimension into variables. Defaults to "var".',
)
@click.option(
    "--output",
    "-o",
    metavar="OUTPUT",
    help="Output path. If omitted, 'INPUT-vars2dim.FORMAT' will be used.",
)
@click.option(
    "--format",
    "-f",
    metavar="FORMAT",
    type=click.Choice(["zarr", "netcdf"]),
    help="Format of the output. If not given, guessed from OUTPUT.",
)
def vars2dim(cube, variable, dim_name, output=None, format=None):
    """Convert cube variables into new dimension.
    Moves all variables of CUBE into a single new variable <var-name>
    with a new dimension DIM-NAME and writes the results to OUTPUT.
    """

    import os

    from xcube.core.dsio import guess_dataset_format, open_dataset, write_dataset
    from xcube.core.vars2dim import vars_to_dim

    if not output:
        dirname = os.path.dirname(cube)
        basename = os.path.basename(cube)
        basename, ext = os.path.splitext(basename)
        output = os.path.join(dirname, basename + "-vars2dim" + ext)

    format_name = format if format else guess_dataset_format(output)

    with open_dataset(input_path=cube) as ds:
        converted_dataset = vars_to_dim(ds, dim_name=dim_name, var_name=variable)
        write_dataset(converted_dataset, output_path=output, format_name=format_name)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import threading
from collections.abc import Sequence
from time import sleep
from timeit import default_timer

import requests

from xcube.util.assertions import assert_given, assert_true
from xcube.util.progress import ProgressObserver, ProgressState

from ...constants import LOG
from .config import CallbackConfig


def _format_time(t):
    """Format seconds into a human readable form.
    Taken form Dask

    >>> _format_time(10.4)
    '10.4s'
    >>> _format_time(1000.4)
    '16min 40.4s'
    """
    m, s = divmod(t, 60)
    h, m = divmod(m, 60)
    if h:
        return f"{h:2.0f}hr {m:2.0f}min {s:4.1f}s"
    elif m:
        return f"{m:2.0f}min {s:4.1f}s"
    else:
        return f"{s:4.1f}s"


class _ThreadedProgressObserver(ProgressObserver):
    """A threaded Progress observer adapted from Dask's ProgressBar class."""

    def __init__(self, minimum: float = 0, dt: float = 1, timeout: float = None):
        """Args:
        dt (float)
        minimum (float)
        """
        super().__init__()
        assert_true(dt >= 0, "The timer's time step must be >=0")
        assert_true(minimum >= 0, "The timer's minimum must be >=0")
        self._running = False
        self._start_time = None
        self._minimum = minimum
        self._dt = dt
        self._width = 100
        self._current_sender = None
        self._state_stack: [ProgressState] = []
        self._timeout = timeout

    def _timer_func(self):
        """Background thread for updating the progress bar"""
        while self._running:
            elapsed = default_timer() - self._start_time
            if elapsed > self._minimum:
                self._update_state(elapsed)
            if self._timeout and elapsed > self._timeout:
                self._running = False
            sleep(self._dt)

    def _start_timer(self):
        self._width = self._state_stack[0].total_work
        self._start_time = default_timer()
        # Start background thread
        self._running = True
        self._timer = threading.Thread(target=self._timer_func)
        self._timer.daemon = True
        self._timer.start()

    def _stop_timer(self, errored):
        self._running = False
        self._timer.join()
        elapsed = default_timer() - self._start_time
        self.last_duration = elapsed
        if elapsed < self._minimum:
            return
        if not errored:
            self.callback(self._current_sender, elapsed, self._state_stack)
        else:
            self._update_state(elapsed)

    def _update_state(self, elapsed):
        if not self._state_stack:
            self.callback(self._current_sender, 0, elapsed)
            return
        state = self._state_stack[0]

        if state.completed_work < state.total_work:
            self.callback(self._current_sender, elapsed, self._state_stack)

    def on_begin(self, state_stack: Sequence[ProgressState]):
        assert_given(state_stack, name="state_stack")
        self._state_stack = state_stack
        self._current_sender = "on_begin"
        if not self._running and len(state_stack) == 1:
            self._start_timer()

    def on_update(self, state_stack: Sequence[ProgressState]):
        assert_given(state_stack, name="state_stack")
        self._state_stack = state_stack
        self._current_sender = "on_update"

    def on_end(self, state_stack: Sequence[ProgressState]):
        assert_given(state_stack, name="state_stack")
        self._state_stack = state_stack
        self._current_sender = "on_end"

        if self._running and len(state_stack) == 1:
            self._stop_timer(False)

    def callback(
        self, sender: str, elapsed: float, state_stack: Sequence[ProgressState]
    ):
        """Args:
            sender
            elapsed
            state_stack

        Returns:

        """


class ApiProgressCallbackObserver(_ThreadedProgressObserver):
    def __init__(
        self,
        callback_config: CallbackConfig,
        minimum: float = 0,
        dt: float = 1,
        timeout: float = False,
    ):
        super().__init__(minimum=minimum, dt=dt, timeout=timeout)
        assert_true(
            callback_config.api_uri and callback_config.access_token,
            "Both, api_uri and access_token must be given.",
        )

        self.callback_config = callback_config

    def callback(
        self, sender: str, elapsed: float, state_stack: Sequence[ProgressState]
    ):
        assert_given(state_stack, "ProgressStates")
        state = state_stack[0]
        callback = {
            "sender": sender,
            "state": {
                "label": state.label,
                "total_work": state.total_work,
                "error": state.exc_info_text or False,
                "progress": state.progress,
                "elapsed": elapsed,
            },
        }
        callback_api_uri = self.callback_config.api_uri
        callback_api_access_token = self.callback_config.access_token
        header = {"Authorization": f"Bearer {callback_api_access_token}"}

        return requests.put(callback_api_uri, json=callback, headers=header)


class TerminalProgressCallbackObserver(_ThreadedProgressObserver):
    def __init__(self, minimum: float = 0, dt: float = 1, timeout: float = False):
        super().__init__(minimum=minimum, dt=dt, timeout=timeout)

    def callback(
        self,
        sender: str,
        elapsed: float,
        state_stack: [ProgressState],
        prt: bool = True,
    ):
        state = state_stack[0]

        bar = "#" * int(state.total_work * state.progress)
        percent = int(100 * state.progress)
        elapsed = _format_time(elapsed)
        msg = "\r{0}: [{1:<{2}}] | {3}% Completed | {4}".format(
            state.label, bar, state.total_work, percent, elapsed
        )

        if prt:
            LOG.info(msg)

        return msg


class ConsoleProgressObserver(ProgressObserver):
    def on_begin(self, state_stack: Sequence[ProgressState]):
        LOG.info(self._format_progress(state_stack, status_label="..."))

    def on_update(self, state_stack: Sequence[ProgressState]):
        LOG.info(self._format_progress(state_stack))

    def on_end(self, state_stack: Sequence[ProgressState]):
        if state_stack[0].exc_info:
            LOG.info(self._format_progress(state_stack, status_label="error!"))
        else:
            LOG.info(self._format_progress(state_stack, status_label="done."))

    @classmethod
    def _format_progress(
        cls, state_stack: Sequence[ProgressState], status_label=None
    ) -> str:
        if status_label:
            state_stack_part = cls._format_state_stack(state_stack[0:-1])
            state_part = cls._format_state(state_stack[-1], marker=status_label)
            return (
                state_part
                if not state_stack_part
                else state_stack_part + ": " + state_part
            )
        else:
            return cls._format_state_stack(state_stack)

    @classmethod
    def _format_state_stack(
        cls, state_stack: Sequence[ProgressState], marker=None
    ) -> str:
        return ": ".join([cls._format_state(s) for s in state_stack])

    @classmethod
    def _format_state(cls, state: ProgressState, marker=None) -> str:
        if marker is None:
            return f"{state.label} - {state.progress:3.1%}"
        else:
            return f"{state.label} - {marker}"



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import jsonschema

from xcube.server.api import ApiError, ApiHandler

from .api import api
from .context import JOB_STATUSES, ComputeContext
from .controllers import get_compute_operation, get_compute_operations

OP_SCHEMA = {
    "type": "object",
    "properties": {
        "operationId": {"type": "string"},
        "parametersSchema": {
            "type": "array",
            "items": {
                "type": "object",
                "description": "JSON Schema for each parameter",
            },
        },
        "description": {"type": "string"},
    },
    "required": ["operationId", "parametersSchema"],
}

OP_LIST_SCHEMA = {
    "type": "object",
    "properties": {
        "operations": {"type": "array", "items": OP_SCHEMA},
    },
    "required": ["operations"],
}

OP_RESPONSES_SCHEMA = {
    "200": {
        "description": "Compute operation details",
        "content": {"application/json": OP_SCHEMA},
    },
    "404": {"description": "An operation with the specified ID was not found."},
}

OP_LIST_RESPONSES_SCHEMA = {
    "200": {
        "description": "Compute operation details",
        "content": {"application/json": {"schema": OP_LIST_SCHEMA}},
    },
    "404": {"description": "An operation with the specified ID was not found."},
}

JOB_STATE_SCHEMA = {
    "type": "object",
    "properties": {
        "status": {"type": "string", "enum": list(JOB_STATUSES)},
        "error": {
            "type": "object",
            "properties": {
                "message": {"type": "string"},
                "type": {"type": "string"},
                "traceback": {"type": "array", "items": {"type": "string"}},
            },
            "required": ["message"],
        },
    },
}

JOB_RESULT_SCHEMA = {
    "type": "object",
    "properties": {
        "datasetId": {"type": "string"},
        "data": {},
    },
}

JOB_SCHEMA = {
    "type": "object",
    "properties": {
        "jobId": {"type": "integer"},
        "state": JOB_STATE_SCHEMA,
        "result": JOB_RESULT_SCHEMA,
    },
    "required": ["jobId", "state"],
}

JOB_LIST_SCHEMA = {
    "type": "object",
    "properties": {
        "jobs": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "jobId": {"type": "integer"},
                    "operationId": {"type": "string"},
                    "status": {"type": "string", "enum": list(JOB_STATUSES)},
                },
                "required": ["jobId", "operationId", "status"],
            },
        },
    },
    "required": ["jobs"],
}

JOB_RESPONSES_SCHEMA = {
    "200": {
        "description": "Compute job details",
        "content": {"application/json": JOB_SCHEMA},
    },
    "404": {"description": "A job with the specified ID was not found."},
}

JOB_LIST_RESPONSES_SCHEMA = {
    "200": {
        "description": "List of compute jobs",
        "content": {"application/json": {"schema": JOB_LIST_SCHEMA}},
    }
}

JOB_REQUEST_SCHEMA = {
    "type": "object",
    "properties": {
        "operationId": {"type": "string"},
        "parameters": {"type": "object", "items": {}},  # Any
        "output": {
            "type": "object",
            "properties": {
                "datasetId": {"type": "string"},
                "title": {"type": "string"},
                "description": {"type": "string"},
                "style": {"type": "string"},
                "colorMappings": {"type": "object"},
            },
            # Doesn't work with swagger:
            # TypeError: t.get is not a function
            # "additionalProperties": False,
        },
        # Doesn't work with swagger:
        # TypeError: t.get is not a function
        # "additionalProperties": False,
    },
    "required": ["operationId"],
    "example": {
        "operationId": "spatial_subset",
        "parameters": {"dataset": "demo", "bbox": [1, 51, 4, 52]},
        "output": {"datasetId": "demo_subset", "title": "My demo subset"},
    },
}

JOB_REQUEST_BODY_SCHEMA = {
    "description": "Compute job request",
    "required": True,
    "content": {"application/json": {"schema": JOB_REQUEST_SCHEMA}},
}

PATH_PARAM_OP_ID = {
    "name": "operationId",
    "in": "path",
    "description": "Operation identifier",
    "schema": {"type": "string"},
}

PATH_PARAM_JOB_ID = {
    "name": "jobId",
    "in": "path",
    "description": "Job identifier",
    "schema": {"type": "integer"},
}


@api.route("/compute/operations")
class ComputeOperationsHandler(ApiHandler[ComputeContext]):
    @api.operation(
        operation_id="getComputeOperations",
        summary="Get available compute operations.",
        responses=OP_LIST_RESPONSES_SCHEMA,
    )
    def get(self):
        self.response.finish(get_compute_operations(self.ctx))


# noinspection PyPep8Naming
@api.route("/compute/operations/{operationId}")
class ComputeOperationHandler(ApiHandler[ComputeContext]):
    @api.operation(
        operation_id="getComputeOperationInfo",
        summary="Get the details of a given compute operation.",
        parameters=[PATH_PARAM_OP_ID],
        responses=OP_RESPONSES_SCHEMA,
    )
    def get(self, operationId):
        self.response.finish(get_compute_operation(self.ctx, operationId))


# noinspection PyPep8Naming
@api.route("/compute/jobs")
class ComputeJobsHandler(ApiHandler[ComputeContext]):
    @api.operation(
        operation_id="getComputeJobs",
        summary="Get recent compute jobs.",
        responses=JOB_LIST_RESPONSES_SCHEMA,
    )
    def get(self):
        self.response.finish(
            {
                "jobs": [
                    {
                        "jobId": job["jobId"],
                        "operationId": job["request"]["operationId"],
                        "status": job["state"]["status"],
                    }
                    for job in self.ctx.jobs.values()
                ]
            }
        )

    @api.operation(
        operation_id="scheduleComputeJob",
        summary="Schedule a new compute job.",
        request_body=JOB_REQUEST_BODY_SCHEMA,
        responses=JOB_RESPONSES_SCHEMA,
    )
    def put(self):
        # TODO: build a more complete schema and publish it as OpenAPI?
        #  Currently, the schema is built for the current request and only
        #  contains (and only needs to contain) a parameter schema for the
        #  requested operation. In xcube’s live OpenAPI specification,
        #  the parameters are just an object with no further constraints.
        #  Ideally we’d want to detail the parameters for the various
        #  available operations in the OpenAPI specification. This can be
        #  done with if-then constructs: see
        #  https://json-schema.org/understanding-json-schema/reference/conditionals.html#if-then-else
        job_request = self.request.json
        basic_schema = self.put.__openapi__["requestBody"]["content"][
            "application/json"
        ]["schema"]
        # noinspection PyProtectedMember,PyUnresolvedReferences
        op_schema = dict(
            properties=dict(
                operationId=dict(enum=list(self.ctx.op_registry.ops.keys())),
                parameters=self.ctx.op_registry.ops[
                    job_request["operationId"]
                ]._op_info.params_schema,
            )
        )
        full_schema = {**basic_schema, **op_schema}
        try:
            jsonschema.validate(job_request, full_schema)
            self.response.finish(self.ctx.schedule_job(job_request))
        except jsonschema.ValidationError as e:
            raise ApiError.BadRequest(message=f"{e.message} at {e.json_path}")


# noinspection PyPep8Naming
@api.route("/compute/jobs/{jobId}")
class ComputeJobHandler(ApiHandler[ComputeContext]):
    @api.operation(
        operation_id="getComputeJob",
        summary="Get details about a compute job.",
        parameters=[PATH_PARAM_JOB_ID],
        responses=JOB_RESPONSES_SCHEMA,
    )
    def get(self, jobId: str):
        job = self.ctx.jobs.get(int(jobId))
        if job is None:
            raise ApiError.NotFound(f"job #{jobId} cannot be found")
        self.response.finish(job)

    @api.operation(
        operation_id="cancelComputeJob",
        summary="Cancel an existing compute job.",
        parameters=[PATH_PARAM_JOB_ID],
        responses=JOB_RESPONSES_SCHEMA,
    )
    def delete(self, jobId: str):
        self.response.finish(self.ctx.cancel_job(int(jobId)))



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import click


# noinspection PyShadowingBuiltins
@click.command(name="dump")
@click.argument("input")
@click.option(
    "--variable",
    "--var",
    metavar="VARIABLE",
    multiple=True,
    help="Name of a variable (multiple allowed).",
)
@click.option(
    "--encoding",
    "-E",
    is_flag=True,
    flag_value=True,
    help="Dump also variable encoding information.",
)
def dump(input, variable, encoding):
    """Dump contents of an input dataset."""
    from xcube.core.dsio import open_dataset
    from xcube.core.dump import dump_dataset

    with open_dataset(input) as ds:
        text = dump_dataset(ds, var_names=variable, show_var_encoding=encoding)
        print(text)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import json
import os
import socket
import threading
from collections.abc import Iterable, Mapping
from pathlib import Path
from typing import Any, Optional, Union

import fsspec
import tornado.ioloop
import xarray as xr

from xcube.constants import LOG
from xcube.core.mldataset import MultiLevelDataset
from xcube.server.server import Server
from xcube.server.webservers.tornado import TornadoFramework
from xcube.webapi.datasets.context import DatasetsContext

# Name of the env var that contains a JupyterLab's base URL.
# If this env var is set, the following _LAB_INFO_FILE will not be used.
# We also imply that a jupyter-server-proxy is installed and enabled.
_LAB_URL_ENV_VAR = "XCUBE_JUPYTER_LAB_URL"

# The following file is generated by the xcube JupyterLab extension
# (xcube-jl-ext) if it is installed and enabled in JupyterLab.
_LAB_INFO_FILE = "~/.xcube/jupyterlab/lab-info.json"


_DEFAULT_MAX_DEPTH = 1


class Viewer:
    """xcube Viewer for Jupyter Notebooks.

    The viewer can be used to visualise and inspect datacubes
    with at least one data variable with dimensions ``["time", "lat", "lon"]``
    or, if a grid mapping is present, with arbitrary ``"time"`` and
    arbitrarily x- and y-dimensions, e.g., ``["time", "y", "x"]`` .

    Add datacubes from instances of ``xarray.Dataset``:

    ```
    viewer = Viewer()
    viewer.add_dataset(dataset)  # can set color styles here too, see doc below
    viewer.show()
    ```

    Display all datasets of formats Zarr, NetCDF, COG/GeoTIFF found in the
    given directories in the local filesystem or in a given S3 bucket:

    ```
    viewer = Viewer(roots=["/eodata/smos/l2", "s3://xcube/examples"])
    viewer.show()
    ```

    The `Viewer` class takes a xcube server configuration as first
    argument. More details regarding configuration parameters are given in the
    `server documentation <https://xcube.readthedocs.io/en/latest/cli/xcube_serve.html>`_.
    The full configuration reference can be generated by executing CLI command
    ``$ xcube serve --show configschema``.

    Args:
        server_config: Server configuration.
            See also output of ``$ xcube serve --show configschema``.
        roots: A path or URL or an iterable of paths or URLs
            that will each be scanned for datasets to be shown in the viewer.
        max_depth: defines the maximum subdirectory depth used to
            search for datasets in case *roots* is given.
    """

    def __init__(
        self,
        server_config: Optional[Mapping[str, Any]] = None,
        roots: Optional[Union[str, Iterable[str]]] = None,
        max_depth: Optional[int] = None,
    ):
        self._server_config, server_url = _get_server_config(
            server_config=server_config, roots=roots, max_depth=max_depth
        )
        self._server_url = server_url
        self._viewer_url = f"{server_url}/viewer/?serverUrl={server_url}"

        # Got trick from
        # https://stackoverflow.com/questions/55201748/running-a-tornado-server-within-a-jupyter-notebook
        self._io_loop = tornado.ioloop.IOLoop()
        thread = threading.Thread(target=self._io_loop.start)
        thread.daemon = True
        thread.start()

        self._server = Server(
            TornadoFramework(io_loop=self._io_loop, shared_io_loop=True),
            config=self._server_config,
        )

        self._io_loop.add_callback(self._server.start)

    @property
    def server_config(self) -> Mapping[str, Any]:
        """The server configuration used by this viewer."""
        return self._server_config

    @property
    def server_url(self):
        """The URL of the server used by this viewer."""
        return self._server_url

    @property
    def viewer_url(self):
        """The URL of this viewer."""
        return self._viewer_url

    @property
    def is_server_running(self) -> bool:
        """Whether the server is running."""
        return self._server is not None

    @property
    def datasets_ctx(self) -> DatasetsContext:
        """Gets the context for the server's "datasets" API."""
        assert self.is_server_running
        return self._server.ctx.get_api_ctx("datasets")

    def stop_server(self):
        """Stops this viewer's server."""
        if self._server is not None:
            # noinspection PyBroadException
            try:
                self._server.stop()
            except BaseException:
                pass
        self._server = None
        self._io_loop = None

    def add_dataset(
        self,
        dataset: Union[xr.Dataset, MultiLevelDataset],
        ds_id: Optional[str] = None,
        title: Optional[str] = None,
        description: Optional[str] = None,
        style: Optional[str] = None,
        color_mappings: dict[str, dict[str, Any]] = None,
    ):
        """Add a dataset to this viewer.

        Args:
            dataset: The dataset to me added. Must be an instance of
                ``xarray.Dataset`` or
                ``xcube.core.mldataset.MultiLevelDataset``.
            ds_id: Optional dataset identifier. If not given, an
                identifier will be generated and returned.
            title: Optional dataset title. Overrides a title given by
                dataset metadata.
            description: Optional dataset description. Overrides a description given by
                dataset metadata.
            style: Optional name of a style that must exist in the
                server configuration.
            color_mappings: Maps a variable name to a specific color
                mapping that is a dictionary comprising a "ValueRange"
                (a pair of numbers) and a "ColorBar" (a matplotlib color
                bar name).

        Returns:
            The dataset identifier.
        """
        if not self._check_server_running():
            return
        return self.datasets_ctx.add_dataset(
            dataset,
            ds_id=ds_id,
            title=title,
            description=description,
            style=style,
            color_mappings=color_mappings,
        )

    def remove_dataset(self, ds_id: str):
        """Remove a dataset from this viewer.

        Args:
            ds_id: The identifier of the dataset to be removed.
        """
        if not self._check_server_running():
            return
        self.datasets_ctx.remove_dataset(ds_id)

    def show(self, width: Union[int, str] = "100%", height: Union[str, int] = 800):
        """Show this viewer as an iframe.
        Intended to be used in a Jupyter notebook.
        If used outside a Jupyter notebook the viewer will be shown
        as a new browser tab.

        Args:
            width: The width of the viewer's iframe.
            height: The height of the viewer's iframe.
        """
        try:
            # noinspection PyPackageRequirements
            from IPython.core.display import HTML

            return HTML(
                f'<iframe src="{self._viewer_url}&compact=1"'
                f' width="{width}"'
                f' height="{height}"'
                f"/>"
            )
        except ImportError as e:
            print(f"Error: {e}; Trying to open Viewer in web browser...")
            # noinspection PyBroadException
            try:
                import webbrowser

                webbrowser.open_new_tab(self.viewer_url)
            except BaseException:
                print("Failed too.")

    def info(self):
        """Output viewer info."""
        # Consider outputting this as HTML if in Notebook
        print(f"Server: {self.server_url}")
        print(f"Viewer: {self.viewer_url}")

    def _check_server_running(self):
        if not self.is_server_running:
            print("Server not running")
        return self.is_server_running


def _get_server_config(
    server_config: Optional[Mapping[str, Any]] = None,
    roots: Optional[Union[str, Iterable[str]]] = None,
    max_depth: Optional[int] = None,
) -> tuple[dict[str, Any], str]:
    server_config = dict(server_config or {})
    max_depth = max_depth or _DEFAULT_MAX_DEPTH

    port = server_config.get("port")
    address = server_config.get("address")

    if port is None:
        port = _find_port()
    if address is None:
        address = "0.0.0.0"

    server_config["port"] = port
    server_config["address"] = address

    server_url, reverse_url_prefix = _get_server_url_and_rev_prefix(port)
    server_config["reverse_url_prefix"] = reverse_url_prefix

    if roots is not None:
        roots = [roots] if isinstance(roots, str) else roots
        config_stores = list(server_config.get("DataStores", []))
        root_stores = _get_data_stores_from_roots(roots, max_depth)
        server_config["DataStores"] = config_stores + root_stores

    return server_config, server_url


def _get_server_url_and_rev_prefix(port: int) -> tuple[str, str]:
    lab_url = os.environ.get(_LAB_URL_ENV_VAR) or None
    has_proxy = lab_url is not None

    if not lab_url:
        lab_info_path = Path(*_LAB_INFO_FILE.split("/")).expanduser()
        if lab_info_path.exists():
            try:
                with lab_info_path.open() as fp:
                    lab_info = json.load(fp)
                lab_url = lab_info["lab_url"]
                has_proxy = lab_info["has_proxy"]
            except (OSError, KeyError):
                LOG.warning(f"Failed loading {lab_info_path}")
                pass

    if lab_url and has_proxy:
        reverse_prefix = f"/proxy/{port}"
        if lab_url.endswith("/"):
            lab_url = lab_url[:-1]
        return lab_url + reverse_prefix, reverse_prefix

    return f"http://localhost:{port}", ""


def _find_port(start: int = 8000, end: Optional[int] = None) -> int:
    """Find a port not in use in range *start* to *end*"""
    end = end if isinstance(end, int) and end >= start else start + 12000
    for port in range(start, end + 1):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            if s.connect_ex(("localhost", port)) != 0:
                return port
    raise RuntimeError("No available port found")


def _get_data_stores_from_roots(
    roots: Iterable[str], max_depth: int
) -> list[dict[str, dict]]:
    extra_data_stores = []
    for index, root in enumerate(roots):
        protocol, path = fsspec.core.split_protocol(root)
        extra_data_stores.append(
            {
                "Identifier": f"_root_{index}",
                "StoreId": protocol or "file",
                "StoreParams": {"root": path, "max_depth": max_depth},
            }
        )
    return extra_data_stores



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from collections.abc import Sequence

import xarray as xr

from xcube.util.progress import observe_progress

from ..config import CubeConfig
from .transformer import TransformedCube


class CubesCombiner:
    def __init__(self, cube_config: CubeConfig):
        self._cube_config = cube_config

    # noinspection PyMethodMayBeStatic
    def combine_cubes(self, t_cubes: Sequence[TransformedCube]) -> TransformedCube:
        cube, gm, _ = t_cubes[0]
        if len(t_cubes) == 1:
            return cube, gm, self._cube_config

        with observe_progress("merging cubes", 1) as observer:
            cube = xr.merge([t_cube[0] for t_cube in t_cubes])
            observer.worked(1)

        return cube, gm, self._cube_config



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import inspect
import threading
from collections.abc import Mapping, Sequence
from typing import Any, Callable, Optional, Tuple, Type, TypeVar, Union

from xcube.util.assertions import assert_true

GET_EXTENSIONS_ATTR_NAME = "get_extensions"

Base = TypeVar("Base")
Ext = TypeVar("Ext")

BaseClass = type[Base]
ExtClass = type[Ext]

ExtClassItem = tuple[str, ExtClass]
ExtInstItem = tuple[str, Ext]


def extend(
    base_class: BaseClass,
    name: str,
    doc: Optional[str] = None,
    class_handler: Union[None, str, Callable[[ExtClassItem], Any]] = None,
    inst_handler: Union[None, str, Callable[[ExtInstItem], Any]] = None,
    ext_args: Optional[Sequence[Any]] = None,
    ext_kwargs: Optional[Mapping[str, Any]] = None,
) -> Callable[[ExtClass], ExtClass]:
    """A class decorator factory that adds an instance property named
    *property_name* to an existing class *base_class*. The value of the
    new property is an instance of the decorated class.

    This decorator can be used to add new functionality to existing
    classes without deriving from the existing class.

    The optional *class_handler* callback function can be used for
    registering the extensions in the base class and/or for
    validating the passed extension name and class.
    Likewise, the optional *inst_handler* callback function can be used for
    registering the extensions in the base class instances and/or for
    validating the passed extension name and instances.

    The decorated class is expected to have a constructor whose first
    positional argument is an instance of *base_class*.
    Other positional and keyword arguments may follow
    and may be passed to this decorator using the *ext_args*
    and *ext_kwargs* keyword arguments.

    Args:
        base_class: The base class to be extended.
        name: The new property's name.
        doc: The new property's docstring, optional.
            If omitted, the docstring of the decorated class, if any, is used.
        class_handler: Optional class handler.
            If given, it must be a callable that is called
            with a tuple comprising extension name and class.
            If this is a string, it is the name of a static or class method
            of the *base_class* which is called with that tuple.
        inst_handler: Optional instance handler.
            If given, it must be a callable that is called
            with a tuple comprising extension name and instance.
            If this is a string, it is the name of an instance method
            of the *base_class* which is called with that tuple.
        ext_args: positional arguments passed
            to the decorated class' constructor
        ext_kwargs: keyword arguments passed
            to the decorated class' constructor

    Returns: a class decorator function
    """
    assert_true(
        inspect.isclass(base_class),
        message=f"base_class must be a class type, but was {type(base_class).__name__}",
    )
    assert_true(
        isinstance(name, str) and name.isidentifier(),
        message=f"name must be a valid identifier, but was {name!r}",
    )
    assert_true(doc is None or isinstance(doc, str), message="doc must be a string")

    _name = "_" + name
    lock = threading.RLock()

    def call_handler(base, handler, ext):
        if isinstance(handler, str):
            getattr(base, handler)((name, ext))
        elif callable(handler):
            handler((name, ext))

    def add_class_property(ext_class: ExtClass) -> ExtClass:
        assert_true(
            inspect.isclass(ext_class),
            message="the extend() decorator can be used with classes only",
        )

        def _get_property(base: Base):
            ext = getattr(base, _name, None)
            if ext is None:
                with lock:
                    ext = getattr(base, _name, None)
                    if ext is None:
                        ext = ext_class(base, *(ext_args or ()), **(ext_kwargs or {}))
                        setattr(base, _name, ext)
                        call_handler(base, inst_handler, ext)
            return ext

        assert_true(
            not hasattr(base_class, name),
            message=f"a property named {name} already"
            f" exists in class {base_class.__name__}",
        )

        setattr(
            base_class,
            name,
            property(
                _get_property, None, None, ext_class.__doc__ if doc is None else doc
            ),
        )

        call_handler(base_class, class_handler, ext_class)

        return ext_class

    return add_class_property



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from typing import Any, Dict, Optional, Tuple

import fsspec
import rasterio
import xarray as xr

from xcube.core.mldataset import LazyMultiLevelDataset, MultiLevelDataset
from xcube.core.store import MULTI_LEVEL_DATASET_TYPE, DataType
from xcube.util.assertions import assert_instance, assert_true
from xcube.util.jsonschema import JsonArraySchema, JsonNumberSchema, JsonObjectSchema

from .dataset import DatasetGeoTiffFsDataAccessor


class GeoTIFFMultiLevelDataset(LazyMultiLevelDataset):
    """A multi-level dataset for GeoTIFF format.

    Args:
        fs: fsspec.AbstractFileSystem object.
        root: Optional root path identifier.
        data_id: dataset identifier.
        open_params: keyword arguments.
    """

    def __init__(
        self,
        fs: fsspec.AbstractFileSystem,
        root: Optional[str],
        data_id: str,
        **open_params: dict[str, Any],
    ):
        super().__init__(ds_id=data_id)
        self._fs = fs
        self._root = root
        self._path = data_id
        self._open_params = open_params
        self._file_url = None

    def _get_overview_count(self):
        with rasterio.open(self._file_url) as rio_dataset:
            overviews = rio_dataset.overviews(1)
        return overviews

    def _get_num_levels_lazily(self) -> int:
        self._file_url = self._get_file_url()
        if isinstance(self._fs, fsspec.AbstractFileSystem):
            with DatasetGeoTiffFsDataAccessor.create_env_session(self._fs):
                overviews = self._get_overview_count()
        else:
            assert_true(self._fs is None, message="invalid type for fs")
        return len(overviews) + 1

    def _get_dataset_lazily(self, index: int, parameters) -> xr.Dataset:
        tile_size = self._open_params.get("tile_size", (512, 512))
        self._file_url = self._get_file_url()
        return DatasetGeoTiffFsDataAccessor.open_dataset(
            self._fs,
            self._file_url,
            tile_size,
            overview_level=index - 1 if index > 0 else None,
        )

    def _get_file_url(self):
        if isinstance(self._fs.protocol, str):
            protocol = self._fs.protocol
        else:
            protocol = self._fs.protocol[0]
        return protocol + "://" + self._path


MULTI_LEVEL_GEOTIFF_OPEN_DATA_PARAMS_SCHEMA = JsonObjectSchema(
    properties=dict(
        tile_size=JsonArraySchema(
            items=(
                JsonNumberSchema(minimum=256, default=512),
                JsonNumberSchema(minimum=256, default=512),
            ),
            default=[512, 512],
        ),
    ),
    additional_properties=False,
)


# noinspection PyAbstractClass
class MultiLevelDatasetGeoTiffFsDataAccessor(DatasetGeoTiffFsDataAccessor):
    """
    Opener/writer extension name: "mldataset:geotiff:<protocol>"
    """

    @classmethod
    def get_data_type(cls) -> DataType:
        return MULTI_LEVEL_DATASET_TYPE

    def get_open_data_params_schema(self, data_id: str = None) -> JsonObjectSchema:
        return MULTI_LEVEL_GEOTIFF_OPEN_DATA_PARAMS_SCHEMA

    def open_data(self, data_id: str, **open_params) -> MultiLevelDataset:
        assert_instance(data_id, str, name="data_id")
        fs, root, open_params = self.load_fs(open_params)
        return GeoTIFFMultiLevelDataset(fs, root, data_id, **open_params)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from typing import Tuple

import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.webapi.compute.op.decorator import op_param, operation


@operation
@op_param(
    "bbox",
    title="Bounding box",
    description="Bounding box using the dataset's CRS coordinates",
    # The rest of the schema is inferred from the function signature.
)
def spatial_subset(
    dataset: xr.Dataset, bbox: tuple[float, float, float, float]
) -> xr.Dataset:
    """Create a spatial subset from given dataset."""
    x1, y1, x2, y2 = bbox
    gm = GridMapping.from_dataset(dataset)
    x_name, y_name = gm.xy_dim_names
    return dataset.sel(
        {
            x_name: slice(x1, x2),
            y_name: slice(y1, y2) if gm.is_j_axis_up else slice(y2, y1),
        }
    )



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import traceback

from xcube.core.normalize import DatasetIsNotACubeError, decode_cube
from xcube.core.store import (
    DATASET_TYPE,
    DataStoreError,
    DataStorePool,
    get_data_store_instance,
    new_data_opener,
)
from xcube.util.assertions import assert_instance
from xcube.util.progress import observe_progress

from ..config import CubeConfig, InputConfig
from ..error import CubeGeneratorError
from .transformer import TransformedCube

# Names of cube configuration parameters that
# are not shared with open parameters.
_STEADY_CUBE_CONFIG_NAMES = {"chunks", "tile_size"}


class CubeOpener:
    def __init__(self, cube_config: CubeConfig, store_pool: DataStorePool = None):
        assert_instance(cube_config, CubeConfig, "cube_config")
        if store_pool is not None:
            assert_instance(store_pool, DataStorePool, "store_pool")
        self._cube_config = cube_config
        self._store_pool = store_pool

    def open_cube(self, input_config: InputConfig) -> TransformedCube:
        cube_config = self._cube_config
        cube_params = cube_config.to_dict()
        opener_id = input_config.opener_id
        store_params = input_config.store_params or {}
        open_params = input_config.open_params or {}

        with observe_progress("reading cube", 3) as observer:
            try:
                if input_config.store_id:
                    store_instance = get_data_store_instance(
                        input_config.store_id,
                        store_params=store_params,
                        store_pool=self._store_pool,
                    )
                    store = store_instance.store
                    if opener_id is None:
                        opener_id = self._get_opener_id(input_config, store)
                    opener = store
                    open_params = dict(open_params)
                    open_params["opener_id"] = opener_id
                else:
                    opener = new_data_opener(opener_id)
                    open_params = dict(open_params)
                    open_params.update(store_params)

                open_params_schema = opener.get_open_data_params_schema(
                    input_config.data_id
                )

                dataset_open_params = {
                    k: v
                    for k, v in cube_params.items()
                    if k in open_params_schema.properties
                }

                observer.worked(1)

                dataset = opener.open_data(
                    input_config.data_id, **open_params, **dataset_open_params
                )
                observer.worked(1)

            except DataStoreError as dse:
                raise CubeGeneratorError(f"{dse}", status_code=400) from dse

            # Turn dataset into cube and grid_mapping
            try:
                cube, gm, _ = decode_cube(dataset, normalize=True)
            except DatasetIsNotACubeError as e:
                raise CubeGeneratorError(f"{e}") from e
            observer.worked(1)

        if dataset_open_params:
            drop_names = [
                k
                for k in dataset_open_params.keys()
                if k not in _STEADY_CUBE_CONFIG_NAMES
            ]
            cube_config = cube_config.drop_props(drop_names)

        return cube, gm, cube_config

    @classmethod
    def _get_opener_id(cls, input_config, store) -> str:
        opener_ids = None
        data_type_names = store.get_data_types_for_data(input_config.data_id)
        for data_type_name in data_type_names:
            if DATASET_TYPE.is_super_type_of(data_type_name):
                opener_ids = store.get_data_opener_ids(
                    data_id=input_config.data_id, data_type=data_type_name
                )
                break
        if not opener_ids:
            raise CubeGeneratorError(
                f"Data store {input_config.store_id!r} does not support datasets",
                status_code=400,
            )
        opener_id = opener_ids[0]
        return opener_id



from typing import Any

import altair as alt
import geopandas as gpd
import pandas as pd
import pyproj
import shapely
import shapely.ops
from shapely.geometry import Point
import xarray as xr

from chartlets import Component, Input, State, Output
from chartlets.components import Box, Button, Typography, Select, VegaChart

from xcube.webapi.viewer.contrib import Panel, get_dataset
from xcube.server.api import Context
from xcube.constants import CRS_CRS84
from xcube.core.extract import get_cube_values_for_points
from xcube.core.gridmapping import GridMapping

panel = Panel(__name__, title="Spectrum View (Demo)", icon="light", position=4)


@panel.layout(
    State("@app", "selectedDatasetId"),
    State("@app", "selectedTimeLabel"),
    State("@app", "selectedPlaceGroup"),
    State("@app", "themeMode"),
)
def render_panel(
    ctx: Context,
    dataset_id: str,
    time_label: str,
    place_group: list[dict[str, Any]],
    theme_mode: str,
) -> Component:

    if theme_mode == "light":
        theme_mode = "default"

    plot = VegaChart(id="plot", chart=None, style={"paddingTop": 6}, theme=theme_mode)

    text = f"{dataset_id} " f"/ {time_label[0:-1]}"
    place_text = Typography(id="text", children=[text], align="center")

    place_names = get_places(ctx, place_group)
    select_places = Select(
        id="select_places",
        label="places (points)",
        value="",
        options=place_names,
    )

    button = Button(id="button", text="Update", style={"maxWidth": 100})

    controls = Box(
        children=[select_places, button],
        style={
            "display": "flex",
            "flexDirection": "row",
            "alignItems": "center",
            "gap": 6,
            "padding": 6,
        },
    )

    control_bar = Box(
        children=[place_text, controls],
        style={
            "display": "flex",
            "flexDirection": "row",
            "alignItems": "center",
            "justifyContent": "space-between",
            "width": "100%",
            "gap": 6,
        },
    )

    return Box(
        children=[
            "Select a map point from the dropdown and press 'Update' "
            "to create a spectrum plot for that point and the selected time.",
            control_bar,
            plot,
        ],
        style={
            "display": "flex",
            "flexDirection": "column",
            "alignItems": "center",
            "width": "100%",
            "height": "100%",
            "gap": 6,
        },
    )


def get_wavelength(
    dataset: xr.Dataset,
    place_group: gpd.GeoDataFrame,
    places: list,
) -> pd.DataFrame:

    grid_mapping = GridMapping.from_dataset(dataset)

    # if place_geometry is not None and not grid_mapping.crs.is_geographic:
    project = pyproj.Transformer.from_crs(
        CRS_CRS84, grid_mapping.crs, always_xy=True
    ).transform

    place_group["geometry"] = place_group["geometry"].apply(
        lambda geom: shapely.ops.transform(project, geom)
    )
    place_group["x"] = place_group["geometry"].apply(
        lambda geom: geom.x if geom else None
    )
    place_group["y"] = place_group["geometry"].apply(
        lambda geom: geom.y if geom else None
    )

    dataset_place = get_cube_values_for_points(dataset, place_group, include_refs=True)

    result = pd.DataFrame()

    for place in places:

        i = (dataset_place.name_ref == place).argmax().item()
        selected_values = (
            dataset_place.drop_vars("geometry_ref")
            .sel(idx=i)
            .compute()
            .to_dict()["data_vars"]
        )

        variables = list(selected_values.keys())
        values = [selected_values[var]["data"] for var in variables]
        wavelengths = [
            dataset_place[var].attrs.get("wavelength", None) for var in variables
        ]

        res = {
            "places": place,
            "variable": variables,
            "reflectance": values,
            "wavelength": wavelengths,
        }

        res = pd.DataFrame(res)
        result = pd.concat([result, res])

    result = result.dropna(subset=["wavelength"])
    return result


# TODO - add selectedDatasetName to Available State Properties
@panel.callback(
    State("@app", "selectedDatasetId"),
    State("@app", "selectedTimeLabel"),
    Input("@app", "selectedTimeLabel"),
    Output("text", "children"),
)
def update_text(
    ctx: Context,
    dataset_id: str,
    time_label: str,
    _time_label: bool | None = None,
) -> list | None:

    text = f"{dataset_id} " f"/ {time_label[0:-1]}"

    return [text]


@panel.callback(
    State("@app", "selectedDatasetId"),
    State("@app", "selectedTimeLabel"),
    State("@app", "selectedPlaceGroup"),
    State("select_places", "value"),
    Input("button", "clicked"),
    Output("plot", "chart"),
)
def update_plot(
    ctx: Context,
    dataset_id: str,
    time_label: str,
    place_group: list[dict[str, Any]],
    place: list,
    _clicked: bool | None = None,
) -> alt.Chart | None:

    if not place_group:
        return None

    if not place:
        return None

    dataset = get_dataset(ctx, dataset_id)

    place_group = gpd.GeoDataFrame(
        [
            {
                "id": feature["id"],
                "name": feature["properties"]["label"],
                "color": feature["properties"]["color"],
                "x": feature["geometry"]["coordinates"][0],
                "y": feature["geometry"]["coordinates"][1],
                "geometry": Point(
                    feature["geometry"]["coordinates"][0],
                    feature["geometry"]["coordinates"][1],
                ),
            }
            for feature in place_group[0]["features"]
            if feature.get("geometry", {}).get("type") == "Point"
        ]
    )

    place_group["time"] = pd.to_datetime(time_label).tz_localize(None)
    place = [place]
    source = get_wavelength(dataset, place_group, place)

    if source is None:
        # TODO: set error message in panel UI
        print("No reflectances found in Variables")
        return None

    chart = (
        alt.Chart(source)
        .mark_line(point=True)
        .encode(
            x="wavelength:Q",
            y="reflectance:Q",
            color="places:N",
            tooltip=["variable", "wavelength", "reflectance"],
        )
    ).properties(width=300, height=200)

    return chart


@panel.callback(
    Input("@app", "selectedPlaceGroup"),
    Output("select_places", "options"),
)
def get_places(
    ctx: Context,
    place_group: list[dict[str, Any]],
) -> list[str]:

    if not place_group:
        return []
    else:
        return [
            feature["properties"]["label"]
            for feature in place_group[0]["features"]
            if feature.get("geometry", {}).get("type") == "Point"
        ]


@panel.callback(
    State("@app", "themeMode"),
    Input("@app", "themeMode"),
    Output("plot", "theme"),
)
def update_theme(
    ctx: Context,
    theme_mode: str,
    _new_theme: bool | None = None,
) -> str:

    if theme_mode == "light":
        theme_mode = "default"

    return theme_mode


# TODO - add selectedDatasetName to Available State Properties
@panel.callback(
    State("@app", "selectedDatasetId"),
    State("@app", "selectedTimeLabel"),
    State("@app", "selectedPlaceGroup"),
    State("select_places", "value"),
    Input("@app", "selectedTimeLabel"),
    Output("plot", "chart"),
)
def update_timestep(
    ctx: Context,
    dataset_id: str,
    time_label: str,
    place_group: list[dict[str, Any]],
    place: list,
    _new_time_label: bool | None = None,
) -> alt.Chart | None:

    return update_plot(ctx, dataset_id, time_label, place_group, place)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import datetime
import fnmatch
import io
import json
import os
import os.path
import string
from collections.abc import Iterable
from typing import Any, Dict, List, Optional, Tuple, Type

import fsspec
import yaml

from xcube.constants import LOG

PRIMITIVE_TYPES = (int, float, str, type(None))

NameAnyDict = dict[str, Any]
NameDictPairList = list[tuple[str, Optional[NameAnyDict]]]


def to_resolved_name_dict_pairs(
    name_dict_pairs: NameDictPairList, container, keep=False
) -> NameDictPairList:
    resolved_pairs = []
    for name, value in name_dict_pairs:
        if "*" in name or "?" in name or "[" in name:
            for resolved_name in container:
                if fnmatch.fnmatch(resolved_name, name):
                    resolved_pairs.append((resolved_name, value))
        elif name in container or keep:
            resolved_pairs.append((name, value))
    return resolved_pairs


def to_name_dict_pairs(iterable: Iterable[Any], default_key=None) -> NameDictPairList:
    return [
        to_name_dict_pair(item, parent=iterable, default_key=default_key)
        for item in iterable
    ]


def to_name_dict_pair(name: Any, parent: Any = None, default_key=None):
    value = None

    if isinstance(name, str):
        try:
            # is key of parent?
            value = parent[name]
        except (KeyError, TypeError, ValueError):
            pass
    else:
        # key = (key, value)?
        try:
            name, value = name
        except (TypeError, ValueError):
            # key = {key: value}?
            try:
                name, value = dict(name).popitem()
            except (TypeError, ValueError, AttributeError, KeyError):
                pass

    if not isinstance(name, str):
        raise ValueError(f"name must be a string")

    if value is None:
        return name, None

    try:
        # noinspection PyUnresolvedReferences
        value.items()
    except AttributeError as e:
        if default_key:
            value = {default_key: value}
        else:
            raise ValueError(f"value of {name!r} must be a dictionary") from e

    return name, value


def flatten_dict(d: dict[str, Any]) -> dict[str, Any]:
    result = dict()
    value = _flatten_dict_value(d, result, None, False)
    if value is not result:
        raise ValueError("input must be a mapping object")
    # noinspection PyTypeChecker
    return result


def _flatten_dict_value(
    value: Any, result: dict[str, Any], parent_name: Optional[str], concat: bool
) -> Any:
    if isinstance(value, datetime.datetime):
        return datetime.datetime.isoformat(value)
    elif isinstance(value, datetime.date):
        return datetime.date.isoformat(value)
    elif isinstance(value, PRIMITIVE_TYPES):
        return value

    try:
        items = value.items()
    except AttributeError:
        items = None

    if items:
        for k, v in items:
            if not isinstance(k, str):
                raise ValueError("all keys must be strings")
            v = _flatten_dict_value(v, result, k, False)
            if v is not result:
                name = k if parent_name is None else parent_name + "_" + k
                if concat and name in result:
                    result[name] = f"{result[name]}, {v}"
                else:
                    result[name] = v

    for e in value:
        _flatten_dict_value(e, result, parent_name, True)

    return result


def merge_config(first_dict: dict, *more_dicts):
    if not more_dicts:
        output_dict = first_dict
    else:
        output_dict = dict(first_dict)
        for d in more_dicts:
            for k, v in d.items():
                if (
                    k in output_dict
                    and isinstance(output_dict[k], dict)
                    and isinstance(v, dict)
                ):
                    v = merge_config(output_dict[k], v)
                output_dict[k] = v
    return output_dict


def load_configs(
    *config_paths: str, exception_type: type[Exception] = ValueError
) -> dict[str, Any]:
    config_dicts = []
    for config_path in config_paths:
        config_dict = load_json_or_yaml_config(
            config_path, exception_type=exception_type
        )
        config_dicts.append(config_dict)
    config = merge_config(*config_dicts)
    return config


def load_json_or_yaml_config(
    config_path: str, exception_type: type[Exception] = ValueError
) -> dict[str, Any]:
    try:
        config_dict = _load_json_or_yaml_config(config_path)
        LOG.info(f"Configuration loaded: {config_path}")
    except FileNotFoundError as e:
        raise exception_type(f"Cannot find configuration {config_path!r}") from e
    except yaml.YAMLError as e:
        raise exception_type(f"YAML in {config_path!r} is invalid: {e}") from e
    except OSError as e:
        raise exception_type(
            f"Cannot load configuration from {config_path!r}: {e}"
        ) from e
    if config_dict is None:
        return {}
    if not isinstance(config_dict, dict):
        raise exception_type(
            f"Invalid configuration format in {config_path!r}: dictionary expected"
        )
    return config_dict


def _load_json_or_yaml_config(config_file: str) -> Any:
    with fsspec.open(config_file, mode="r") as fp:
        file_content = fp.read()
    template = string.Template(file_content)
    file_content = template.safe_substitute(os.environ)
    with io.StringIO(file_content) as fp:
        if config_file.endswith(".json"):
            return json.load(fp)
        else:
            return yaml.safe_load(fp)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import os

default_header = """Copyright (c) 2018-2025 by xcube team and contributors
Permissions are hereby granted under the terms of the MIT License:
https://opensource.org/licenses/MIT."""


def update_copyright(
    dir_path: str = ".",
    header: str | None = None,
    prefix: str = "#",
    ext: str = ".py",
):
    header_lines = list(
        map(lambda line: f"{prefix} {line}\n", (header or default_header).split("\n"))
    )
    for root, dirs, files in os.walk(dir_path):
        for filename in files:
            if filename.endswith(ext) and filename != "update-copyright.py":
                _update_copyright(os.path.join(root, filename), header_lines, prefix)


def _update_copyright(file_path: str, header_lines: list[str], prefix: str):
    with open(file_path) as f:
        lines = f.readlines()

    old_header_size = 0
    is_copyright = False
    for line in lines:
        if line.startswith(prefix):
            is_copyright = is_copyright or "copyright" in line.lower()
            old_header_size += 1
        else:
            break

    if is_copyright or old_header_size == 0:
        with open(file_path, "w") as f:
            f.writelines(header_lines)
            if old_header_size == 0:
                f.write("\n")
            f.writelines(lines[old_header_size:])
        if old_header_size == 0:
            print("Added missing CR header: ", file_path)
    else:
        print("Not a CR header: ", file_path)


update_copyright("../xcube")
update_copyright("../test")



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.
from collections.abc import Sequence

from xcube.util.config import flatten_dict, load_configs, to_name_dict_pairs


def get_config_dict(
    config_files: Sequence[str] = None,
    input_paths: Sequence[str] = None,
    input_processor_name: str = None,
    output_path: str = None,
    output_writer_name: str = None,
    output_size: str = None,
    output_region: str = None,
    output_variables: str = None,
    output_resampling: str = None,
    append_mode: bool = True,
    profile_mode: bool = False,
    no_sort_mode: bool = False,
):
    """Get a configuration dictionary from given (command-line) arguments.

    Returns:
        Configuration dictionary

    Raises: OSError, ValueError
    """

    config = load_configs(*config_files) if config_files else {}

    # preserve backward compatibility for old names
    if "input_processor" in config:
        config["input_processor_name"] = config.pop("input_processor")
    if "output_writer" in config:
        config["output_writer_name"] = config.pop("output_writer")

    # Overwrite current configuration by cli arguments
    if input_paths is not None and "input_paths" not in config:
        if len(input_paths) == 1 and input_paths[0].endswith(".txt"):
            with open(input_paths[0]) as input_txt:
                input_paths = input_txt.readlines()
            config["input_paths"] = [x.strip() for x in input_paths]
        else:
            config["input_paths"] = input_paths

    if input_processor_name is not None and "input_processor_name" not in config:
        config["input_processor_name"] = input_processor_name

    if output_path is not None and "output_path" not in config:
        config["output_path"] = output_path

    if output_writer_name is not None and "output_writer_name" not in config:
        config["output_writer_name"] = output_writer_name

    if output_resampling is not None and "output_resampling" not in config:
        config["output_resampling"] = output_resampling

    if output_size is not None:
        try:
            output_size = list(map(lambda c: int(c), output_size.split(",")))
        except ValueError:
            output_size = None
        if output_size is None or len(output_size) != 2:
            raise ValueError(
                f"output_size must have the form <width>,<height>,"
                f" where both values must be positive integer numbers"
            )
        config["output_size"] = output_size

    if output_region is not None:
        try:
            output_region = list(map(lambda c: float(c), output_region.split(",")))
        except ValueError:
            output_region = None
        if output_region is None or len(output_region) != 4:
            raise ValueError(
                f"output_region must have the form <lon_min>,<lat_min>,<lon_max>,<lat_max>,"
                f" where all four numbers must be floating point numbers in degrees"
            )
        config["output_region"] = output_region

    if output_variables is not None:
        output_variables = list(map(lambda c: c.strip(), output_variables.split(",")))
        if output_variables == [""] or any(
            [var_name == "" for var_name in output_variables]
        ):
            raise ValueError(
                "output_variables must be a list of existing variable names"
            )
        config["output_variables"] = output_variables

    if profile_mode is not None and config.get("profile_mode") is None:
        config["profile_mode"] = profile_mode

    if append_mode is not None and config.get("append_mode") is None:
        config["append_mode"] = append_mode

    if no_sort_mode is not None and config.get("no_sort_mode") is None:
        config["no_sort_mode"] = no_sort_mode

    processed_variables = config.get("processed_variables")
    if processed_variables:
        config["processed_variables"] = to_name_dict_pairs(processed_variables)

    output_variables = config.get("output_variables")
    if output_variables:
        config["output_variables"] = to_name_dict_pairs(
            output_variables, default_key="name"
        )

    output_metadata = config.get("output_metadata")
    if output_metadata:
        config["output_metadata"] = flatten_dict(output_metadata)

    return config



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.


import collections.abc
from collections.abc import Iterable, Iterator, KeysView
from logging import Logger
from typing import Optional

import zarr.storage

from xcube.constants import LOG
from xcube.util.assertions import assert_instance
from xcube.util.perf import measure_time_cm


class LoggingZarrStore(zarr.storage.BaseStore):
    """A Zarr Store that logs all method calls on another store *other*
    including execution time.
    """

    def __init__(
        self,
        other: collections.abc.MutableMapping,
        logger: Logger = LOG,
        name: Optional[str] = None,
    ):
        assert_instance(other, collections.abc.MutableMapping)
        self._other = other
        self._measure_time = measure_time_cm(logger=logger)
        self._name = name or "chunk_store"
        if hasattr(other, "listdir"):
            setattr(self, "listdir", self.__listdir)
        if hasattr(other, "getsize"):
            setattr(self, "getsize", self.__getsize)

    def __listdir(self, key: str) -> Iterable[str]:
        with self._measure_time(f"{self._name}.listdir({key!r})"):
            # noinspection PyUnresolvedReferences
            return self._other.listdir(key)

    def __getsize(self, key: str) -> int:
        with self._measure_time(f"{self._name}.getsize({key!r})"):
            # noinspection PyUnresolvedReferences
            return self._other.getsize(key)

    def keys(self) -> KeysView[str]:
        with self._measure_time(f"{self._name}.keys()"):
            # noinspection PyTypeChecker
            return self._other.keys()

    def __iter__(self) -> Iterator[str]:
        with self._measure_time(f"{self._name}.__iter__()"):
            return self._other.__iter__()

    def __len__(self) -> int:
        with self._measure_time(f"{self._name}.__len__()"):
            return self._other.__len__()

    def __contains__(self, key) -> bool:
        with self._measure_time(f"{self._name}.__contains__({key!r})"):
            return self._other.__contains__(key)

    def __getitem__(self, key: str) -> bytes:
        with self._measure_time(f"{self._name}.__getitem__({key!r})"):
            return self._other.__getitem__(key)

    def __setitem__(self, key: str, value: bytes) -> None:
        with self._measure_time(f"{self._name}.__setitem__({key!r}, <value>)"):
            return self._other.__setitem__(key, value)

    def __delitem__(self, key: str) -> None:
        with self._measure_time(f"{self._name}.__delitem__({key!r})"):
            return self._other.__delitem__(key)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from collections.abc import Container
from typing import Any, Union

_DEFAULT_NAME = "value"


def assert_not_none(
    value: Any, name: str = None, exception_type: type[Exception] = ValueError
):
    """Assert *value* is not None.
    Otherwise, raise *exception_type*.

    Args:
        value: The value to test.
        name: Name of a variable that holds *value*.
        exception_type: The exception type. Default is ``ValueError``.
    """
    if value is None:
        raise exception_type(f"{name or _DEFAULT_NAME} must not be None")


def assert_given(
    value: Any, name: str = None, exception_type: type[Exception] = ValueError
):
    """Assert *value* is not False when converted into a Boolean value.
    Otherwise, raise *exception_type*.

    Args:
        value: The value to test.
        name: Name of a variable that holds *value*.
        exception_type: The exception type. Default is ``ValueError``.
    """
    if not value:
        raise exception_type(f"{name or _DEFAULT_NAME} must be given")


def assert_instance(
    value: Any,
    dtype: Union[type, tuple[type, ...]],
    name: str = None,
    exception_type: type[Exception] = TypeError,
):
    """Assert *value* is an instance of data type *dtype*.
    Otherwise, raise *exception_type*.

    Args:
        value: The value to test.
        dtype: A type or tuple of types.
        name: Name of a variable that holds *value*.
        exception_type: The exception type. Default is ``TypeError``.
    """
    if not isinstance(value, dtype):
        raise exception_type(
            f"{name or _DEFAULT_NAME} must be an instance of {dtype}, was {type(value)}"
        )


def assert_subclass(
    value: Any,
    cls: Union[type, tuple[type, ...]],
    name: str = None,
    exception_type: type[Exception] = TypeError,
):
    """Assert *value* is a subclass of class *cls*.
    Otherwise, raise *exception_type*.

    Args:
        value: The value to test.
        cls: A class or tuple of classes.
        name: Name of a variable that holds *value*.
        exception_type: The exception type. Default is ``TypeError``.
    """
    if not issubclass(value, cls):
        raise exception_type(
            f"{name or _DEFAULT_NAME} must be a subclass of {cls}, was {value}"
        )


def assert_in(
    value: Any,
    container: Container,
    name: str = None,
    exception_type: type[Exception] = ValueError,
):
    """Assert *value* is a member of *container*.
    Otherwise, raise *exception_type*.

    Args:
        value: The value to test for membership.
        container: The container.
        name: Name of a variable that holds *value*.
        exception_type: The exception type. Default is ``ValueError``.
    """
    if value not in container:
        raise exception_type(f"{name or _DEFAULT_NAME} must be one of {container}")


def assert_true(value: Any, message: str, exception_type: type[Exception] = ValueError):
    """Assert *value* is true after conversion into a Boolean value.
    Otherwise, raise *exception_type*.

    Args:
        value: The value to test.
        message: The error message used if the assertion fails.
        exception_type: The exception type. Default is ``ValueError``.
    """
    if not value:
        raise exception_type(message)


def assert_false(
    value: Any, message: str, exception_type: type[Exception] = ValueError
):
    """Assert *value* is false after conversion into a Boolean value.
    Otherwise, raise *exception_type*.

    Args:
        value: The value to test.
        message: The error message used if the assertion fails.
        exception_type: The exception type. Default is ``ValueError``.
    """
    if value:
        raise exception_type(message)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from abc import ABC, abstractmethod
from typing import Tuple

import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.util.progress import observe_progress

from ..config import CubeConfig
from .helpers import is_empty_cube, strip_cube

TransformedCube = tuple[xr.Dataset, GridMapping, CubeConfig]


class CubeTransformer(ABC):
    @abstractmethod
    def transform_cube(
        self, cube: xr.Dataset, gm: GridMapping, cube_config: CubeConfig
    ) -> TransformedCube:
        """
        Transform given *cube*, grid mapping *gm*, and cube
        configuration *cube_config* into a potentially new cube,
        a new grid mapping, and less restrictive cube configuration
        *cube_config*.

        The latter is achieved by returning a derived *cube_config* where
        all properties that have been "consumed" by this transformer
        are removed. See :meth:`CubeConfig.drop_property`.
        """


class CubeIdentity(CubeTransformer):
    def transform_cube(
        self, cube: xr.Dataset, gm: GridMapping, cube_config: CubeConfig
    ) -> TransformedCube:
        """Return *cube*, grid mapping *gm*, and parameters without change."""
        return cube, gm, cube_config


def transform_cube(
    t_cube: TransformedCube, transformer: CubeTransformer, label: str = ""
) -> TransformedCube:
    empty_cube = is_empty_cube(t_cube[0])
    identity = isinstance(transformer, CubeIdentity)
    if not label:
        label = f"{type(transformer).__name__}"
    if identity:
        label += " (step not applicable)"
    elif empty_cube:
        label += " (step not applicable, empty cube)"

    with observe_progress(label, 1) as progress:
        if not (identity or empty_cube):
            t_cube = transformer.transform_cube(*t_cube)
            t_cube = strip_cube(t_cube[0]), t_cube[1], t_cube[2]
        progress.worked(1)

    return t_cube



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import tempfile
from collections.abc import MutableMapping
from typing import Dict, Tuple, Union

import numpy as np
import xarray as xr
import zarr

from xcube.core.chunk import chunk_dataset
from xcube.core.unchunk import unchunk_dataset

DEFAULT_TIME_EPS = np.array(1000 * 1000, dtype="timedelta64[ns]")


def find_time_slice(
    store: Union[str, MutableMapping],
    time_stamp: Union[np.datetime64, np.ndarray],
    time_eps: np.timedelta64 = DEFAULT_TIME_EPS,
) -> tuple[int, str]:
    """Find time index and update mode for *time_stamp* in
    Zarr dataset given by *store*.

    Args:
        store: A zarr store.
        time_stamp: Time stamp to find index for.
        time_eps: Time epsilon for equality comparison,
            defaults to 1 millisecond.
    Returns:
        A tuple (time_index, 'insert') or (time_index, 'replace')
        if an index was found, (-1, 'create') or (-1, 'append')
        otherwise.
    """
    try:
        cube = xr.open_zarr(store)
    except (FileNotFoundError, ValueError):
        # FileNotFoundError is raised as by Zarr since 2.13,
        # before GroupNotFoundError (extends ValueError) was raised.
        # Keep ValueError for backward compatibility.
        try:
            cube = xr.open_dataset(store)
        except (FileNotFoundError, ValueError):
            # If the zarr directory does not exist, open_dataset raises a
            # FileNotFoundError (with xarray <= 0.17.0) or a ValueError
            # (with xarray 0.18.0).
            return -1, "create"

    # TODO (forman): optimise following naive search by bi-sectioning or so
    for i in range(cube.time.size):
        time = cube.time[i]
        if abs(time_stamp - time) < time_eps:
            return i, "replace"
        if time_stamp < time:
            return i, "insert"

    return -1, "append"


def append_time_slice(
    store: Union[str, MutableMapping],
    time_slice: xr.Dataset,
    chunk_sizes: dict[str, int] = None,
):
    """Append time slice to existing zarr dataset.

    Args:
        store: A zarr store.
        time_slice: Time slice to insert
        chunk_sizes: desired chunk sizes
    """
    if chunk_sizes:
        time_slice = chunk_dataset(time_slice, chunk_sizes, format_name="zarr")

    # Unfortunately time_slice.to_zarr(store, mode='a', append_dim='time') will replace global attributes of store
    # with attributes of time_slice (xarray bug?), which are usually empty in our case.
    # Hence, we must save our old attributes in a copy of time_slice.
    ds = zarr.open_group(store, mode="r")
    time_slice = time_slice.copy()
    time_slice.attrs.update(ds.attrs)
    if "coordinates" in time_slice.attrs:
        # Remove 'coordinates', otherwise we get
        # ValueError: cannot serialize coordinates because the global attribute 'coordinates' already exists
        # from next time_slice.to_zarr(...) call.
        time_slice.attrs.pop("coordinates")

    time_slice.to_zarr(store, mode="a", append_dim="time", consolidated=True)

    unchunk_dataset(store, coords_only=True)


def insert_time_slice(
    store: Union[str, MutableMapping],
    insert_index: int,
    time_slice: xr.Dataset,
    chunk_sizes: dict[str, int] = None,
):
    """Insert time slice into existing zarr dataset.

    Args:
        store: A zarr store.
        insert_index: Time index
        time_slice: Time slice to insert
        chunk_sizes: desired chunk sizes
    """
    update_time_slice(
        store, insert_index, time_slice, "insert", chunk_sizes=chunk_sizes
    )


def replace_time_slice(
    store: Union[str, MutableMapping],
    insert_index: int,
    time_slice: xr.Dataset,
    chunk_sizes: dict[str, int] = None,
):
    """Replace time slice in existing zarr dataset.

    Args:
        store: A zarr store.
        insert_index: Time index
        time_slice: Time slice to insert
        chunk_sizes: desired chunk sizes
    """
    update_time_slice(
        store, insert_index, time_slice, "replace", chunk_sizes=chunk_sizes
    )


def update_time_slice(
    store: Union[str, MutableMapping],
    insert_index: int,
    time_slice: xr.Dataset,
    mode: str,
    chunk_sizes: dict[str, int] = None,
):
    """Update existing zarr dataset by new time slice.

    Args:
        store: A zarr store.
        insert_index: Time index
        time_slice: Time slice to insert
        mode: Update mode, 'insert' or 'replace'
        chunk_sizes: desired chunk sizes
    """

    if mode not in ("insert", "replace"):
        raise ValueError(f"illegal mode value: {mode!r}")

    insert_mode = mode == "insert"

    time_var_names = []
    encoding = {}
    with xr.open_zarr(store) as cube:
        for var_name in cube.variables:
            var = cube[var_name]
            if var.ndim >= 1 and "time" in var.dims:
                if var.dims[0] != "time":
                    raise ValueError(
                        f"dimension 'time' of variable {var_name!r} must be first dimension"
                    )
                time_var_names.append(var_name)
                enc = dict(cube[var_name].encoding)
                # xarray 0.17+ supports engine preferred chunks if exposed by the backend
                # zarr does that, but when we use the new 'preferred_chunks' when writing to zarr
                # it raises and says, 'preferred_chunks' is an unsupported encoding
                if "preferred_chunks" in enc:
                    del enc["preferred_chunks"]
                encoding[var_name] = enc

    if chunk_sizes:
        time_slice = chunk_dataset(time_slice, chunk_sizes, format_name="zarr")
    temp_dir = tempfile.TemporaryDirectory(prefix="xcube-time-slice-", suffix=".zarr")
    time_slice.to_zarr(temp_dir.name, encoding=encoding)
    slice_root_group = zarr.open(temp_dir.name, mode="r")
    slice_arrays = dict(slice_root_group.arrays())

    cube_root_group = zarr.open(store, mode="r+")
    for var_name, var_array in cube_root_group.arrays():
        if var_name in time_var_names:
            slice_array = slice_arrays[var_name]
            if insert_mode:
                # Add one empty time step
                empty = zarr.creation.empty(slice_array.shape, dtype=var_array.dtype)
                var_array.append(empty, axis=0)
                # Shift contents
                var_array[insert_index + 1 :, ...] = var_array[insert_index:-1, ...]
            # Replace slice
            var_array[insert_index, ...] = slice_array[0]

    unchunk_dataset(store, coords_only=True)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import collections.abc
import fnmatch
from collections.abc import Hashable, Mapping
from typing import List, Optional, Union

import dask.array as da
import numpy as np
import xarray as xr
from scipy import stats

from xcube.util.assertions import assert_in, assert_instance, assert_true

AGG_METHODS = "auto", "first", "min", "max", "mean", "median", "mode"
DEFAULT_INT_AGG_METHOD = "first"
DEFAULT_FLOAT_AGG_METHOD = "mean"

AggMethod = Union[None, str]
AggMethods = Union[AggMethod, Mapping[str, AggMethod]]


def subsample_dataset(
    dataset: xr.Dataset,
    step: int,
    xy_dim_names: Optional[tuple[str, str]] = None,
    agg_methods: Optional[AggMethods] = None,
) -> xr.Dataset:
    """Subsample *dataset* with given integer subsampling *step*.
    Only data variables with spatial dimensions given by
    *xy_dim_names* are subsampled.

    Args:
        dataset: the dataset providing the variables
        step: the integer subsampling step size in pixels in
            the x and y directions.
            For aggregation methods other than "first" it defines the
            window size for the aggregation.
        xy_dim_names: the spatial dimension names
        agg_methods: Optional aggregation methods.
            May be given as string or as mapping from variable name pattern
            to aggregation method. Valid aggregation methods are
            "auto", "first", "min", "max", "mean", "median", and "mode".
            If "auto", the default, "first" is used for integer variables
            and "mean" for floating point variables.
    Returns:
        The subsampled dataset or a tuple comprising the
        subsampled dataset and the effective aggregation methods.
    """
    assert_instance(dataset, xr.Dataset, name="dataset")
    assert_instance(step, int, name="step")
    assert_valid_agg_methods(agg_methods)

    x_name, y_name = xy_dim_names or ("y", "x")

    agg_methods = get_dataset_agg_methods(
        dataset, xy_dim_names=xy_dim_names, agg_methods=agg_methods
    )

    new_data_vars = dict()
    new_coords = None  # used to collect coordinates from coarsen
    for var_name, var in dataset.data_vars.items():
        if var_name in agg_methods:
            agg_method = agg_methods[var_name]
            if agg_method == "first":
                slices = get_variable_subsampling_slices(
                    var, step, xy_dim_names=xy_dim_names
                )
                assert slices is not None
                new_var = var[slices]
            else:
                if agg_method == "mode":
                    new_var: xr.DataArray = _agg_mode(var, x_name, y_name, step)
                else:
                    new_var: xr.DataArray = _agg_builtin(
                        var, x_name, y_name, step, agg_method
                    )
                if new_var.dtype != var.dtype:
                    # We don't want, e.g. "mean", to turn data
                    # from dtype unit16 into float64
                    new_var = new_var.astype(var.dtype)
                new_var.attrs.update(var.attrs)
                new_var.encoding.update(var.encoding)
                # coarsen() recomputes spatial coordinates.
                # Collect them, so we can later apply them to the
                # variables that are subsampled by "first"
                # (= slice selection).
                if new_coords is None:
                    new_coords = dict(new_var.coords)
                else:
                    new_coords.update(new_var.coords)
        else:
            new_var = var

        new_data_vars[var_name] = new_var

    if not new_data_vars:
        return dataset

    if new_coords:
        # Make sure all variables use the same modified
        # spatial coordinates from coarsen
        new_data_vars = {
            k: v.assign_coords({d: new_coords[d] for d in v.dims if d in new_coords})
            for k, v in new_data_vars.items()
        }

    return xr.Dataset(data_vars=new_data_vars, attrs=dataset.attrs)


def _agg_mode(var: xr.DataArray, x_name: str, y_name: str, step: int) -> xr.DataArray:
    var_coarsen = _coarsen(var, x_name, y_name, step)
    drop_axis = _get_drop_axis(var, x_name, y_name, step)

    def _scipy_mode(x, axis, **kwargs):
        return stats.mode(x, axis, nan_policy="omit", **kwargs).mode

    def _mode_dask(x, axis, **kwargs):
        return x.map_blocks(
            _scipy_mode, axis=axis, dtype=x.dtype, drop_axis=drop_axis, **kwargs
        )

    if isinstance(var.data, da.Array):
        return var_coarsen.reduce(_mode_dask)
    else:
        return var_coarsen.reduce(_scipy_mode)


def _agg_builtin(
    var: xr.DataArray, x_name: str, y_name: str, step: int, agg_method: str
) -> xr.DataArray:
    var_coarsen = _coarsen(var, x_name, y_name, step)
    return getattr(var_coarsen, agg_method)()


def _coarsen(var: xr.DataArray, x_name: str, y_name: str, step: int):
    dim = dict()
    if x_name in var.dims:
        dim[x_name] = step
    if y_name in var.dims:
        dim[y_name] = step
    return var.coarsen(dim=dim, boundary="pad", coord_func="min")


def _get_drop_axis(var: xr.DataArray, x_name: str, y_name: str, step: int) -> List[int]:
    """
    This function serves to determine the indexes of the dimensions of a coarsened
    array that will not be included in the array after reduction. These dimensions
    are the indexes following the indexes of the spatial dimensions from the original
    array, so first these indexes are determined, then they are shifted appropriately.
    """
    drop_axis = []
    if x_name in var.dims:
        drop_axis.append(var.dims.index(x_name))
    if y_name in var.dims:
        drop_axis.append(var.dims.index(y_name))
    if len(drop_axis) == 1:
        drop_axis[0] += 1
    elif len(drop_axis) == 2:
        # The latter index must be shifted by two positions
        if drop_axis[0] > drop_axis[1]:
            drop_axis[0] += 2
            drop_axis[1] += 1
        else:
            drop_axis[0] += 1
            drop_axis[1] += 2
    return drop_axis


def get_dataset_agg_methods(
    dataset: xr.Dataset,
    xy_dim_names: Optional[tuple[str, str]] = None,
    agg_methods: Optional[AggMethods] = None,
):
    assert_instance(dataset, xr.Dataset, name="dataset")
    assert_valid_agg_methods(agg_methods)

    x_name, y_name = xy_dim_names or ("y", "x")

    dataset_agg_methods = dict()

    for var_name, var in dataset.data_vars.items():
        if x_name in var.dims or y_name in var.dims:
            agg_method = find_agg_method(agg_methods, var_name, var.dtype)
            dataset_agg_methods[str(var_name)] = agg_method

    return dataset_agg_methods


def assert_valid_agg_methods(agg_methods: Optional[AggMethods]):
    """Assert that the given *agg_methods* are valid."""
    assert_instance(
        agg_methods, (type(None), str, collections.abc.Mapping), name="agg_methods"
    )
    if isinstance(agg_methods, str):
        assert_in(agg_methods, AGG_METHODS, name="agg_methods")
    elif agg_methods is not None:
        enum = (None, *AGG_METHODS)
        for k, v in agg_methods.items():
            assert_true(
                isinstance(k, str), message="keys in agg_methods must be strings"
            )
            assert_true(
                v in enum, message=f"values in agg_methods must be one of {enum}"
            )


def find_agg_method(
    agg_methods: AggMethods, var_name: Hashable, var_dtype: np.dtype
) -> str:
    """Find aggregation method in *agg_methods*
    for given *var_name* and *var_dtype*.
    """
    assert_valid_agg_methods(agg_methods)
    if isinstance(agg_methods, str) and agg_methods != "auto":
        return agg_methods
    if isinstance(agg_methods, collections.abc.Mapping):
        for var_name_pat, agg_method in agg_methods.items():
            if var_name == var_name_pat or fnmatch.fnmatch(str(var_name), var_name_pat):
                if agg_method in (None, "auto"):
                    break
                return agg_method
    # here: agg_method is either None or 'auto'
    if np.issubdtype(var_dtype, np.integer):
        return "first"
    else:
        return "mean"


_FULL_SLICE = slice(None, None, None)


def get_variable_subsampling_slices(
    variable: xr.DataArray, step: int, xy_dim_names: Optional[tuple[str, str]] = None
) -> Optional[tuple[slice, ...]]:
    """Compute subsampling array slices for *variable*.

    Args:
        variable: the dataset providing the variables
        step: the integer subsampling step
        xy_dim_names: the spatial dimension names

    Returns:
        A tuple that comprises array slices for all variable dimensions.
        The slices for spatial dimensions are subsampled according
        to *step*, while other slices are cover every array item.
        Returns `None`, if *variable* does not contain spatial
        dimensions at all.
    """
    assert_instance(variable, xr.DataArray, name="variable")
    assert_instance(step, int, name="step")
    x_dim_name, y_dim_name = xy_dim_names or ("x", "y")

    var_index = None
    for index, dim_name in enumerate(variable.dims):
        if dim_name == x_dim_name or dim_name == y_dim_name:
            if var_index is None:
                var_index = index * [_FULL_SLICE]
            var_index.append(slice(None, None, step))
        elif var_index is not None:
            var_index.append(_FULL_SLICE)
    return tuple(var_index) if var_index is not None else None



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from .context import VarExprContext
from .error import VarExprError
from .helpers import split_var_assignment
from .varexpr import VarExpr, evaluate, parse



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from typing import Any, Dict, List, Optional, Union

from xcube.util.jsonschema import (
    JsonArraySchema,
    JsonIntegerSchema,
    JsonNumberSchema,
    JsonObject,
    JsonObjectSchema,
    JsonStringSchema,
)

from ..response import CubeGeneratorResult, CubeInfo, make_cube_generator_result_class


class CubeGeneratorToken(JsonObject):
    def __init__(self, access_token: str, token_type: str):
        self.access_token = access_token
        self.token_type = token_type

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                access_token=JsonStringSchema(min_length=1),
                token_type=JsonStringSchema(min_length=1),
            ),
            required=["access_token", "token_type"],
            additional_properties=False,
            factory=cls,
        )

    @classmethod
    def from_dict(cls, value: dict) -> "CubeGeneratorToken":
        return cls.get_schema().from_instance(value)


class CubeGeneratorJobStatus(JsonObject):
    # noinspection PyUnusedLocal
    def __init__(
        self,
        succeeded: int = None,
        failed: int = None,
        active: int = None,
        start_time: str = None,
        completion_time: str = None,
        conditions: list[dict[str, Any]] = None,
        **additional_properties,
    ):
        self.succeeded: Optional[int] = succeeded
        self.failed: Optional[int] = failed
        self.active: Optional[int] = active
        self.start_time: Optional[str] = start_time
        self.completion_time: Optional[str] = completion_time
        self.conditions: Optional[dict[str, Any]] = conditions
        self.additional_properties: dict[str, Any] = additional_properties

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                succeeded=JsonIntegerSchema(nullable=True),
                failed=JsonIntegerSchema(nullable=True),
                active=JsonIntegerSchema(nullable=True),
                start_time=JsonStringSchema(nullable=True),
                completion_time=JsonStringSchema(nullable=True),
                conditions=JsonArraySchema(
                    items=JsonObjectSchema(additional_properties=True), nullable=True
                ),
            ),
            additional_properties=True,
            factory=cls,
        )

    @classmethod
    def from_dict(cls, value: dict) -> "CubeGeneratorJobStatus":
        return cls.get_schema().from_instance(value)


class CubeGeneratorProgressState(JsonObject):
    """Current progress state of the remote generator."""

    def __init__(
        self,
        progress: float,
        # worked: Union[int, float],
        total_work: Union[int, float],
        **additional_properties,
    ):
        self.progress: float = float(progress)
        # self.worked: float = float(worked)
        self.total_work: float = float(total_work)
        self.additional_properties = additional_properties

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                progress=JsonNumberSchema(minimum=0.0),
                # worked=JsonNumberSchema(minimum=0),
                total_work=JsonNumberSchema(exclusive_minimum=0),
            ),
            required=[
                "progress",
                # 'worked',
                "total_work",
            ],
            additional_properties=True,
            factory=cls,
        )

    @classmethod
    def from_dict(cls, value: dict) -> "CubeGeneratorProgressState":
        return cls.get_schema().from_instance(value)


class CubeGeneratorProgress(JsonObject):
    """Current progress of the remote generator."""

    def __init__(self, sender: str, state: CubeGeneratorProgressState):
        self.sender: str = sender
        self.state: CubeGeneratorProgressState = state

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                sender=JsonStringSchema(),
                state=CubeGeneratorProgressState.get_schema(),
            ),
            required=[
                "sender",
                "state",
            ],
            additional_properties=False,
            factory=cls,
        )

    @classmethod
    def from_dict(cls, value: dict) -> "CubeGeneratorProgress":
        return cls.get_schema().from_instance(value)


class CubeGeneratorState(JsonObject):
    """Current state of the remote generator."""

    def __init__(
        self,
        job_id: str,
        job_status: CubeGeneratorJobStatus,
        job_result: Optional[CubeGeneratorResult] = None,
        output: Optional[list[str]] = None,
        progress: Optional[list[CubeGeneratorProgress]] = None,
        **additional_properties,
    ):
        self.job_id = job_id
        self.job_status = job_status
        self.job_result = job_result
        self.output = output
        self.progress = progress
        self.additional_properties = additional_properties

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                job_id=JsonStringSchema(min_length=1),
                job_status=CubeGeneratorJobStatus.get_schema(),
                job_result=CubeGeneratorResult.get_schema(),
                output=JsonArraySchema(items=JsonStringSchema(), nullable=True),
                progress=JsonArraySchema(
                    items=CubeGeneratorProgress.get_schema(), nullable=True
                ),
            ),
            required=["job_id", "job_status"],
            additional_properties=True,
            factory=cls,
        )

    @classmethod
    def from_dict(cls, value: dict) -> "CubeGeneratorState":
        return cls.get_schema().from_instance(value)


class CostEstimation(JsonObject):
    def __init__(self, required: int, available: int = None, limit: int = None):
        self.available: int = available
        self.limit: Optional[int] = limit
        self.required: Optional[int] = required

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                required=JsonIntegerSchema(minimum=0),
                available=JsonIntegerSchema(minimum=0),
                limit=JsonIntegerSchema(minimum=0),
            ),
            required=["required"],
            additional_properties=False,
            factory=cls,
        )

    @classmethod
    def from_dict(cls, value: dict) -> "CostEstimation":
        return cls.get_schema().from_instance(value)


class CubeInfoWithCosts(CubeInfo):
    def __init__(self, cost_estimation: CostEstimation, **kwargs):
        super().__init__(**kwargs)
        self.cost_estimation: CostEstimation = cost_estimation

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        schema = super().get_schema()
        schema.properties.update(cost_estimation=CostEstimation.get_schema())
        schema.required.append("cost_estimation")
        schema.factory = cls
        return schema


CubeInfoWithCostsResult = make_cube_generator_result_class(CubeInfoWithCosts)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from xcube.server.api import Api

from .config import AuthConfig
from .context import AuthContext

api = Api(
    "auth",
    description="xcube Auth API",
    config_schema=AuthConfig.get_schema(),
    create_ctx=AuthContext,
)
